"""
This file contains the logic of the Gear Solver.

The algorithm requires one Fight Object that will be used to compute the DPS. The algorithm will use
the pre-baked simulation in order to quickly simulate the DPS given some gear configuration.
One list of Gear from which to search, the space of materias that are available, etc.

For now the BiS Solver will only work if the Fight has one player variable, meaning it can only solve the BiS for one 

"""

from ffxivcalc.GearSolver.Gear import GearSet, MateriaGenerator, GearType, StatType
from ffxivcalc.Jobs.PlayerEnum import RoleEnum, JobEnum
from ffxivcalc.helperCode.Progress import ProgressBar
from ffxivcalc.helperCode.exceptions import InvalidFoodSpace, InvalidGearSpace, InvalidMateriaSpace, InvalidFunctionParameter, MultiValuedWeaponDelay
from math import floor
from copy import deepcopy
import os
import logging
main_logging = logging.getLogger("ffxivcalc")
solver_logging = main_logging.getChild("Solver")


def getBaseStat(IsTank=False):
    """
    This function returns a base stat dictionnary
    """

    return {
        "MainStat" : 450 if not IsTank else 390,
        "WD" : 0,
        "Det" : 390,
        "Ten" : 400,
        "SS" : 400,
        "SkS" : 400,
        "Crit" : 400,
        "DH" : 400,
        "Piety" : 390
    }  

def getGearDPSValue(Fight, gearSet : GearSet, PlayerIndex : int,PlayerID : int, n : int =10000):
    """
    This function returns the expected damage of a gear set given a fight
    Fight -> Fight object
    gearSet : GearSet -> Gear set to be tested
    PlayerIndex : int -> index of the player to test the gear set on
    PlayerID : int -> ID of the player
    n : int -> Number of times to run random simulations.
    """
                             # Computes the PreBakedAction and asks the fight object to remember those actions for the player with the given ID.
                             # The simulated player will be given base stats.
    player = Fight.PlayerList[PlayerIndex]
    IsTank = player.RoleEnum == RoleEnum.Tank 
    IsCaster = player.RoleEnum == RoleEnum.Caster or player.RoleEnum == RoleEnum.Healer
    JobMod = player.JobMod # Level 90 jobmod value, specific to each job

    player.Stat = getBaseStat(IsTank=IsTank)
    GearStat = gearSet.GetGearSetStat(IsTank=IsTank)
    GearStat = {'MainStat': 3378, 'WD': 132, 'Det': 1885, 'Ten': 400, 'SS': 400, 'SkS': 400, 'Crit': 2598, 'DH': 1252, 'Piety': 390}
    player.Stat["SS" if IsCaster else "SkS"] = GearStat["SS" if IsCaster else "SkS"]
    Fight.SavePreBakedAction = True
                             # The playerIndex value is the ID value here.
    Fight.PlayerIDSavePreBakedAction = PlayerID
    Fight.SimulateFight(0.01, 500, False, n=0,PPSGraph=False)
    player.DamageInstanceList = []

    #gearSet, x , y = materiaBisSolverV3(gearSet, MateriaGenerator(18,36), [0,1,2], Fight, JobMod, IsTank, IsCaster, 0, "exp", 0, mendSpellSpeed=False, minSPDValue=400, maxSPDValue=500, oversaturationIterationsPostGear=1, findOptMateriaGearBF=True,swapDHDetBeforeSpeed=False)

    f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
    ExpectedDamage, randomDamageDict, duration, potency = Fight.SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=n,getInfo = True)

    print(gearSet)
    print("Expected : " + str(ExpectedDamage))
    print("Random : " + str(randomDamageDict))
    print("Duration : " + str(duration))
    print("Potency : " + str(potency))
    print("f_WD : " + str(f_WD) + " f_DET : " + str(f_DET) + " f_TEN : " + str(f_TEN) + " f_SPD : " + str(f_SPD) + " f_CritRate : " + str(f_CritRate) + " f_CritMult : " + str(f_CritMult) + " f_DH : " + str(f_DH))

    return ExpectedDamage, randomDamageDict

def computeDamageValue(GearStat : dict, JobMod : int, IsTank : bool, IsCaster : bool):
    """
    this function computes all the damage values given a stat dict.
    GearStat : dict -> Dict containing the value of all stats.
    JobMod : int -> JobMod value of the player.
    IsTank : bool -> True if the player is a tank
    IsCaster : bool -> True if the player is a caster (Magical Ranged or Healer)
    """
    levelMod = 1900
    baseMain = 390  
    baseSub = 400# Level 90 LevelMod values

    f_WD = (GearStat["WD"]+floor(baseMain*JobMod/1000))/100 # Necessary to check if its not 0 since etro only returns the damage multiplier.
    f_DET = floor(1000+floor(140*(GearStat["Det"]-baseMain)/levelMod))/1000# Determination damage
    if IsTank : f_TEN = (1000+floor(100*(GearStat["Ten"]-baseSub)/levelMod))/1000 # Tenacity damage, 1 for non-tank player
    else : f_TEN = 1 # if non-tank
    f_SPD = (1000+floor(130*((GearStat["SS"] if IsCaster else GearStat["SkS"])-baseSub)/levelMod))/1000 # Used only for dots
    f_CritRate = floor((200*(GearStat["Crit"]-baseSub)/levelMod+50))/1000 # Crit rate in decimal
    f_CritMult = (floor(200*(GearStat["Crit"]-baseSub)/levelMod+400))/1000 # Crit Damage multiplier
    f_DH = floor(550*(GearStat["DH"]-baseSub)/levelMod)/1000 # DH rate in decimal
    DHAuto = floor(140*(GearStat["DH"]-baseSub)/levelMod)/1000 # DH bonus when auto crit/DH
    return f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto

def computeGCDTimer(speedStatValue : int, subGCDHasteAmount : int) -> (str, str):
    """This function returns the GCD timer with and without haste.

    Args:
        speedStatValue (int): Value of the speed stat
        subGCDHasteAmount (int) : Value of the haste the player might get from abilities.
                                  This is relevant to computing the faster GCD timer which
                                  might be different even if the non-hasted GCD timer is the same.
                                  Relevant to jobs such as BLM, NIN, SAM, WHM and BRD.

    Returns:
        str: GCD timer rounded down to 2 decimals as a string.
    """
    gcdReductionValue = (1000 - floor(130 * (speedStatValue-400) / 1900))/1000
    gcdTimer = str(floor(floor(floor((2500 * gcdReductionValue)))/10)/100)
    if subGCDHasteAmount == 0 : return gcdTimer, gcdTimer
    hastedGCDTimer = str(floor(floor(floor((2500 * gcdReductionValue)) * (100 - subGCDHasteAmount)/100)/10)/100)
    return gcdTimer, hastedGCDTimer

def findGCDTimerRange(minSPDValue : int, maxSPDValue : int, subGCDHasteAmount : int = 0) -> dict:
    """This function finds all possible GCD tiers for speed values going from [minSPDValue, maxSPDValue].
    It returns a dictionnary whos key is a tuple (gcdTimer, hastedGCDTimer). If subGCDHasteAmount if 0 then
    the key is (gcdTimer, gcdTimer). ie -> (2.42, 2.42) if subGCDHasteAmount is 0.
    Args:
        minSPDValue (int): Minimal speed value
        maxSPDValue (int): Maximal speed value
        subGCDHasteAmount (int) : Value of the haste the player might get from abilities.
                                  This is relevant to computing the faster GCD timer which
                                  might be different even if the non-hasted GCD timer is the same.
                                  Relevant to jobs such as BLM, NIN, SAM, WHM and BRD.
    """

    gcdTierList = {}
    gcdTimerDone = []

                             # Finding all values assuming the gear set can full of meld (36*22)=792
    trialSPDValue = max(400, minSPDValue - (792+103))

    while trialSPDValue <= maxSPDValue:
        gcdTimerTuple = computeGCDTimer(trialSPDValue, subGCDHasteAmount)
        if not gcdTimerTuple in gcdTimerDone : gcdTierList[gcdTimerTuple] = trialSPDValue
        gcdTimerDone.append(gcdTimerTuple)
        trialSPDValue += 10

    return gcdTierList

def BiSSolver(Fight, GearSpace : dict, MateriaSpace : list, FoodSpace : list, PercentileToOpt : list = ["exp", "99", "90", "75", "50"],
              materiaDepthSearchIterator : int = 1, randomIteration : int = 10000, oddMateriaValue : int = 18, evenMateriaValue : int = 36,
              PlayerIndex : int = 0, PlayerID : int = 1, mendSpellSpeed : bool = False, maxSPDValue : int = 5000, minSPDValue : int = 0, useNewAlgo : bool = False, oversaturationIterationsPreGear : int = 0,
              oversaturationIterationsPostGear : int = 0, findOptMateriaGearBF : bool = False, swapDHDetBeforeSpeed : bool = True, minPiety : int = 390, gcdTimerSpecificActionList : dict = None,
              saveAsFile : bool = True, showBar : bool = True, loadingBarBuffer = None,returnGearSet : bool = True):
    """
    Finds the BiS of the player given a Gear search space and a Fight. The Solver will output to a file named
    bisSolver[Job]Result[number].txt with all the relevant information and returns the gearSets. The solver outputs the best Expected Damage GearSet as well as
    the best GearSet for the 50th, 75th, 90th and 99th DPS. The solver requires at least one piece of every possible slots.
    Note that is is possible to give a "prebaked" BiS by only giving 1 choice for some pieces and that it is also possible to
    add materias to the pieces in the GearSpace before running the solver. 
    Note that since the percentile's BiS are using random damage this is only an approximation and might miss the actual best in slot. So it 
    is recommended to run this solver multiple times or with very high value of randomIteration. 

    It is also recommend to have useNewAlgo set to true in order to use the faster and more efficient algorithm.
    If you are solving for the Expected Damage BiS it is also recommend to have findOptMateriaGearBF set to True in order to try
    every possible gear set as otherwise the best meld is only found once a gear set is found.

    Fight -> Fight object.
    GearSpace : dict -> Dictionnary filled with the different gear pieces the algorithm can search through. Must have at least one of each gear types
    Materiaspace : list -> List of all the stats the solver will look through when optimizing melds. Must be at least 3 materias
    FoodSpace : list -> List of all food to look into. Must be non-empty
    PercentileToOpt : list -> List of all gearset to optimize. Exp means to optimize expected damage.
    materiaDepthSearchIterator : int -> Depth for which the Materia optimization searches into per step.
    randomIteration : int -> Number of times the solver will simulate random runs. Must be at least 100
    oddMateriaValue : int -> Stat gained from odd Materia
    evenMateriaValue : int -> Stat gained from even Materia
    PlayerIndex : int -> Index of the player for which the user wants to optimize the gearset. Must be the index of the player in the Fight.PlayerList.
    PlayerID : int -> ID of the player for which we want to optimize.
    mendSpellSpeed : bool -> If True the solver will look at Spell Speed value for speed.
    maxSPDValue : int -> Max Spell Speed or skill Speed value. Every gear set above that value will be discarded.
    minSPDValue : int -> Min Spell Speed or skill Speed value. Every gear set under that value will be discarded.
    useNewAlgo : bool -> If set to true will use V2 of materiaBiSSolver
    oversaturationIterationsPreGear : int -> Number of times the algorithm will oversaturate gear before looking for best gear set
    oversaturationIterationsPostGear : int -> Number of times the algorithm will oversaturate gear before looking for best materias
    findOptMateriaGearBF : bool -> If true solver will find best gearset/food/melding using given algorithm. Only recommended for Expected.
    swapDHDetBeforeSpeed : bool -> If True, the solver will swap DH and Det before swapping melds with Speed materias. If False it swaps after.
    minPiety : int -> Minimum required Piety value for the set. By default set to 400.
    gcdTimerSpecificActionList : dict -> Dictionary with key (gcdTimer, hastedGCDTimer) where the key maps to a list of actions
                                         to perform for the given gcdTimer and hastedGCDTimer. If is empty we ignore and only use
                                         the action list present in the fight object. Recommended to use the findGCDTimerRange() function
                                         with required minSPDValue and maxSPDValue in order to get an accurate dictionary. The mapping of gcd Tier
                                         does not have to be exhaustive. If no key is found it will use the rotation of the Fight object instead.
    saveAsFile : bool -> If true saves the result in a file.
    showBar : bool -> If true shows loading bar. If false doesn't but still updates the loading bar's memory
    loadingBarBuffer : dict -> This dictionnary will be given the progress bar's adress at the key 'pb'. Can be used to access the PB.
    returnGearSet : bool = True -> If true returns the gear set,random,text. If false returns the expectedDPS,text
    """

                             # Checking the validity of the given search space and some other parameters.
    if materiaDepthSearchIterator <= 0: raise InvalidFunctionParameter("BisSolver", "materiaDepthSearchIterator", "Must be higher than 1.")
    if randomIteration < 100 : raise InvalidFunctionParameter("BisSolver", "randomIteration", "Must be higher than or 100")
    if PlayerIndex < 0 or PlayerIndex > len(Fight.PlayerList) - 1: raise InvalidFunctionParameter("BisSolver", "PlayerIndex", "Invalid index value")
    if useNewAlgo and oversaturationIterationsPreGear + oversaturationIterationsPostGear < 1: raise InvalidFunctionParameter("BiSSolver", "OverSaturationIteration", "If using newAlgo, must oversaturate gearset at least once. (oversaturationIterationsPreGear + oversaturationIterationsPostGear >= 1)")
    if useNewAlgo and (StatType.SS in MateriaSpace or StatType.SkS in MateriaSpace) : raise InvalidFunctionParameter("BiSSolver", "MateriaSpace", "MateriaSpace cannot invclue Speed Materias while using V2")
    if findOptMateriaGearBF and PercentileToOpt != ["exp"] : raise InvalidFunctionParameter("BiSSolver", "PercentileToOpt", "Can only optimize Expected if findOptMateriaGearBF is True")
    if minSPDValue > maxSPDValue : raise InvalidFunctionParameter("BiSSolver", "Limitations to Speed Values", "minSPDValue cannot be bigger than maxSPDValue")
    expectedGearSpaceKeys = ["WEAPON", "HEAD", "BODY", "HANDS", "LEGS", "FEET", "EARRINGS", "NECKLACE", "BRACELETS", "LRING", "RING"]
    for key in expectedGearSpaceKeys:
        if not (key in GearSpace.keys()) or len(GearSpace[key]) == 0: raise InvalidGearSpace(key)
    if len(MateriaSpace) < 3 : raise InvalidMateriaSpace
    if len(FoodSpace) == 0 : raise InvalidFoodSpace

                             # Computes the PreBakedAction and asks the fight object to remember those actions for the player with the given ID.
                             # The simulated player will be given base stats.
    

                             # Making deep copy of GearSpace so we do not modify the object insides the space
    newGearSpace = {key : [deepcopy(element) for element in GearSpace[key]] for key in GearSpace.keys()}
    GearSpace = newGearSpace

    IsTank = Fight.PlayerList[PlayerIndex].RoleEnum == RoleEnum.Tank
    IsCaster = Fight.PlayerList[PlayerIndex].RoleEnum == RoleEnum.Caster or Fight.PlayerList[PlayerIndex].RoleEnum == RoleEnum.Healer
    JobMod = Fight.PlayerList[PlayerIndex].JobMod # Level 90 jobmod value, specific to each job
    Fight.PlayerList[PlayerIndex].Stat = getBaseStat(IsTank=IsTank)
    Fight.SavePreBakedAction = True
                             # The playerIndex value is the ID value here.
    Fight.PlayerIDSavePreBakedAction = PlayerID

                             # Finding haste amount if any.
    hasteAmount = 0
    match Fight.PlayerList[PlayerIndex].JobEnum:
                             # Assume max haste amount in order to find sub GCD tier.
        case JobEnum.BlackMage : hasteAmount = 15
        case JobEnum.WhiteMage : hasteAmount = 20
        case JobEnum.Samurai : hasteAmount = 13
        case JobEnum.Monk : hasteAmount = 20
        case JobEnum.Bard : hasteAmount = 20
        case JobEnum.Astrologian : hasteAmount = 10
        case JobEnum.Ninja : hasteAmount = 15

                             # Getting all possible gcdTimers from the given range of speed value. Will simulate prebakedsimulation for all of them.
    gcdTimerDict = findGCDTimerRange(minSPDValue, maxSPDValue,subGCDHasteAmount=hasteAmount)
    solver_logging.warning("Computed GCD timer : " + str(gcdTimerDict))
    gcdTimerProgress = ProgressBar.init(len(gcdTimerDict.keys()), "Prebaking GCD tier",showBar=showBar, extraBuffer=loadingBarBuffer)

                             # Getting weaponDelay value. If more than one weaponDelay value returns an error as the solver currently cannot handle more than
                             # one different value (because we are prebaking each fight). It would be possible but would need to prebake each fight also depending
                             # on the weaponDelay value TODO
    weaponDelay = GearSpace["WEAPON"][0].getWeaponDelay() # -> Note that this list is non empty at this point since we checked before hand
    for weapon in GearSpace["WEAPON"]:
        if weaponDelay != weapon.getWeaponDelay():
            # If not equal either means two different weapon delay which is currently not possible (see above)
            raise MultiValuedWeaponDelay(weaponDelay, weapon.getWeaponDelay())
        
    Fight.PlayerList[PlayerIndex].setBasedWeaponDelay(weaponDelay)
                             # Setting base weaponDelay

                             # This dictionnary contains all Fight object. The key is (gcdTimer, hastedGCDTimer)
    preBakedFightGCDTierList = {}
    for key in gcdTimerDict:
        Fight.PlayerList[PlayerIndex].Stat['SS' if IsCaster else "SkS"] = gcdTimerDict[key]
        preBakedFightGCDTierList[key] = Fight.deepCopy()
        

                             # If a specific gcd timer rotation is given then we swap the ActionSet for the gccd specific one.
        if gcdTimerSpecificActionList != None and (key in gcdTimerSpecificActionList.keys()): 
            preBakedFightGCDTierList[key].PlayerList[PlayerIndex].ActionSet = deepcopy(gcdTimerSpecificActionList[key])

        preBakedFightGCDTierList[key].SimulateFight(0.01, 500, False, n=0,PPSGraph=False, showProgress=False,computeGraph=False)
        next(gcdTimerProgress)

    matGen = MateriaGenerator(oddMateriaValue, evenMateriaValue)
    newGearSet = GearSet()
    optimalGearSet = GearSet()
    optimalRandomGearSet = {key : [0,GearSet()] for key in PercentileToOpt}
    optimalRandomGearSetMateria = {key : [0,GearSet(),{}] for key in PercentileToOpt}

                            # Removing exp from optimalRandom.
    if "exp" in optimalRandomGearSet.keys(): 
        optimalRandomGearSet.pop("exp")
        optimalRandomGearSetMateria.pop("exp")

    for key in GearSpace.keys():
        for gear in GearSpace[key]:
            if gear.getIgnoreOptimize() : continue
            for type in MateriaSpace:
                for j in range(oversaturationIterationsPreGear):
                    if type != 3: 
                        for i in range(gear.getNumberPossibleMateria(type, matGen)) : gear.forceAddMateria(matGen.GenerateMateria(type))
                    
    curOptimalDPS = 0
                             # Finding total number of possibilities for ProgressBar
    total = 1
    for key in GearSpace:
        total *= len(GearSpace[key])

    gearBFpB = ProgressBar.init(total, "Finding Best Gear Set",showBar=showBar, extraBuffer=loadingBarBuffer)
                             # Need at least one of each gear piece.
    for Weapon in GearSpace["WEAPON"]:
        newGearSet.AddGear(Weapon)
        for Head in GearSpace["HEAD"]:
            newGearSet.AddGear(Head)
            for Body in GearSpace["BODY"]:
                newGearSet.AddGear(Body)
                for Hands in GearSpace["HANDS"]:
                    newGearSet.AddGear(Hands)
                    for Legs in GearSpace["LEGS"]:
                        newGearSet.AddGear(Legs)
                        for Feet in GearSpace["FEET"]:
                            newGearSet.AddGear(Feet)
                            for Earrings in GearSpace["EARRINGS"]:
                                newGearSet.AddGear(Earrings)
                                for Necklace in GearSpace["NECKLACE"]:
                                    newGearSet.AddGear(Necklace)
                                    for Bracelets in GearSpace["BRACELETS"]:
                                        newGearSet.AddGear(Bracelets)
                                        for LRing in GearSpace["LRING"]:
                                            newGearSet.AddGear(LRing)
                                            for Ring in GearSpace["RING"]:
                                                newGearSet.AddGear(Ring)

                                                GearStat = newGearSet.GetGearSetStat(IsTank=IsTank)
                                                if GearStat["SS" if mendSpellSpeed else "SkS"] > maxSPDValue : continue

                                                curBestFoodExpectedDPS = 0
                                                curBestFoodExpectedSet = None
                                                curBestFoodRandomDPS = {key : [0,None,{}] for key in optimalRandomGearSetMateria}


                                                                        # Adding here so I don't forget. SpS/SkS are not usually as high as the other
                                                                        # stats here because we do not oversaturate with Speed melds. THis could
                                                                        # affect the buff given by food. Should add an option to add all the SpS/SkS bonus
                                                                        # from food to see if this affects the final results.
                                                for food in FoodSpace:
                                                    trialSet = deepcopy(newGearSet)
                                                    foodPercentBuff = food.getPercentStatBonus("SS" if mendSpellSpeed else "SkS")
                                                    foodMaxBuff = food.getLimitStatBonus("SS" if mendSpellSpeed else "SkS")
                                                    GearStat = trialSet.GetGearSetStat(IsTank=IsTank)
                                                    spdStat = GearStat["SS" if mendSpellSpeed else "SkS"] 
                                                    trialSetMinSPDValue = min(int(spdStat * foodPercentBuff),foodMaxBuff) + spdStat
                                                    if trialSetMinSPDValue > maxSPDValue : continue
                                                                         # Will test for SpS/SkS to see if it can fall within the accepted minSPDValue.
                                                                         # Checking if 0 before so we can speedup if it will work for sure.

                                                    
                                                    gearSetMaxSPDValue = ((GearStat["SS"] if mendSpellSpeed else GearStat["SkS"]) + trialSet.getMateriaTypeLimit((StatType.SS if mendSpellSpeed else StatType.SkS), matGen) * matGen.EvenValue)
                                                    gearSetMaxSPDValueFood = min(int(foodPercentBuff * gearSetMaxSPDValue),foodMaxBuff) + gearSetMaxSPDValue
                                                    canReachMinSPD = (minSPDValue == 400) or gearSetMaxSPDValueFood >= minSPDValue

                                                    solver_logging.warning("foodMaxBuff : " + str(foodMaxBuff) + "foodPercentBuff : " + str(foodPercentBuff) + "gearSetMaxSPDValue : " + str(gearSetMaxSPDValue) + " gearSetMaxSPDValueFood : " + str(gearSetMaxSPDValueFood) + " canReachMinSPD :" + str(canReachMinSPD))


                                                    if not canReachMinSPD:
                                                        continue
                                                                         # Will find optimal meld with food
                                                    trialSet.addFood(food)
                                                    solver_logging.warning(trialSet)
                                                    if findOptMateriaGearBF: 
                                                        trialSet, exp, ra = materiaBisSolverV3(trialSet, matGen, MateriaSpace, preBakedFightGCDTierList, hasteAmount, Fight.PlayerList[PlayerIndex].JobMod, IsTank, IsCaster, PlayerIndex,
                                                                                               "exp",0,mendSpellSpeed,minSPDValue=minSPDValue,maxSPDValue=maxSPDValue,oversaturationIterationsPostGear=oversaturationIterationsPostGear,
                                                                                               findOptMateriaGearBF=findOptMateriaGearBF,swapDHDetBeforeSpeed=swapDHDetBeforeSpeed,minPiety=minPiety)

                                                    GearStat = trialSet.GetGearSetStat(IsTank=IsTank)

                                                                    # This is a lazy attempt at fixing the issue, but it might work
                                                                    # Note that is issue only really comes when minPiety and
                                                                    # minSPDValue are both high.
                                                    if minPiety > 390 and GearStat["Piety"] < minPiety : continue

                                                    if not ((mendSpellSpeed and GearStat["SS"] > maxSPDValue) or (not mendSpellSpeed and GearStat["SkS"] > maxSPDValue)):

                                                        JobMod = Fight.PlayerList[PlayerIndex].JobMod # Level 90 jobmod value, specific to each job
                                
                                                        f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                                                        gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
                                                        ExpectedDamage, randomDamageDict = preBakedFightGCDTierList[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, 
                                                                                                                                                    f_CritMult, f_DH, DHAuto, n=randomIteration)

                                                        if curBestFoodExpectedDPS <= ExpectedDamage:
                                                            curBestFoodExpectedSet = deepcopy(trialSet)
                                                            curBestFoodExpectedDPS = ExpectedDamage

                                                        for percentile in curBestFoodRandomDPS:
                                                            if curBestFoodRandomDPS[percentile][0] <= randomDamageDict[percentile]:
                                                                curBestFoodRandomDPS[percentile][0] = randomDamageDict[percentile]
                                                                curBestFoodRandomDPS[percentile][1] = deepcopy(trialSet)
                                                                curBestFoodRandomDPS[percentile][2] = deepcopy(randomDamageDict)

                                                if curOptimalDPS < curBestFoodExpectedDPS:
                                                    optimalGearSet = deepcopy(curBestFoodExpectedSet)
                                                    curOptimalDPS = curBestFoodExpectedDPS

                                                for percentile in optimalRandomGearSet:
                                                    if optimalRandomGearSet[percentile][0] <= curBestFoodRandomDPS[percentile][0]:
                                                        optimalRandomGearSet[percentile][0] = curBestFoodRandomDPS[percentile][0]
                                                        optimalRandomGearSet[percentile][1] = deepcopy(curBestFoodRandomDPS[percentile][1])
                                                next(gearBFpB)        

                             # Only do optimal meld if not already found
    if not findOptMateriaGearBF:
        for percentile in PercentileToOpt:
            if percentile == "exp" : 
                solver_logging.warning("Optimal oversaturated gear set : \n" + (str(optimalGearSet)) + "\nDamage : " + str(curOptimalDPS))
            else : solver_logging.warning("Optimal oversaturated gear set : \n" + (str(optimalRandomGearSet[percentile][1])) + "\nDamage : " + str(optimalRandomGearSet[percentile][0]))
                                # Will now find optimal Meld for the optimal sets that were found.
        if useNewAlgo:
            print("Using BF up-down")
            if "exp" in PercentileToOpt : 
                print("Optimizing Best Expected BiS materia")
                optimalGearSet, curMax, curRandom = materiaBisSolverV3(optimalGearSet, matGen, MateriaSpace, preBakedFightGCDTierList, hasteAmount, Fight.PlayerList[PlayerIndex].JobMod, IsTank, IsCaster, PlayerIndex, "exp",0,mendSpellSpeed,minSPDValue=minSPDValue,maxSPDValue=maxSPDValue,oversaturationIterationsPostGear=oversaturationIterationsPostGear,findOptMateriaGearBF=findOptMateriaGearBF, swapDHDetBeforeSpeed=swapDHDetBeforeSpeed,minPiety=minPiety)
                                    # Will now optimize the random BiS. Every percentile's gearset is optimized by using
                                    # the value of the DPS as their percentile. So the 90th percentile BiS is chosen using the
                                    # materia arrangement that maximizes the 90th percentile DPS.
            for percentile in optimalRandomGearSet:
                if percentile != "exp" : 
                    print("Optimizing " + percentile + "th percentile BiS")
                    optimalRandomGearSetMateria[percentile][1], optimalRandomGearSetMateria[percentile][0], curRandom = materiaBisSolverV3(optimalRandomGearSet[percentile][1], matGen, MateriaSpace, preBakedFightGCDTierList, hasteAmount, Fight.PlayerList[PlayerIndex].JobMod, IsTank, IsCaster, PlayerIndex, percentile,randomIteration,mendSpellSpeed,minSPDValue=minSPDValue,maxSPDValue=maxSPDValue, oversaturationIterationsPostGear=oversaturationIterationsPostGear,swapDHDetBeforeSpeed=swapDHDetBeforeSpeed,minPiety=minPiety)
                    optimalRandomGearSetMateria[percentile][2] = deepcopy(curRandom)
        else:
            print("Using BF down-up")
            depth = materiaDepthSearchIterator 
            limit = optimalGearSet.getMateriaLimit()  # Number of materia the gearset can support
            if materiaDepthSearchIterator > limit: raise InvalidFunctionParameter("BisSolver", "materiaDepthSearchIterator", "Must be lower than the limit of the gear set")

            counter = 0              # Materia counter
            curMax = 0               # cur Max DPS
            curRandom = {}           # DPS percentiles of current best
            if "exp" in PercentileToOpt : print("Optimizing Best Expected BiS materia")
            while True:
                if not "exp" in PercentileToOpt: break
                print("Progress : " + str(counter)+"/"+str(limit))
                if counter + materiaDepthSearchIterator > limit:
                    d = limit - counter
                else:
                    d = materiaDepthSearchIterator
                curBest, curMax, curRandom = materiaBisSolverV3(optimalGearSet, matGen, MateriaSpace, d, preBakedFightGCDTierList, Fight.PlayerList[PlayerIndex].JobMod, IsTank, IsCaster, PlayerIndex, "exp",0,mendSpellSpeed,maxSPDValue=maxSPDValue)
                                    # Taking a copy of the found best and incrementing materia counter
                optimalGearSet = deepcopy(curBest)
                counter += materiaDepthSearchIterator
                if counter >= limit : break
                                    # Will now optimize the random BiS. Every percentile's gearset is optimized by using
                                    # the value of the DPS as their percentile. So the 90th percentile BiS is chosen using the
                                    # materia arrangement that maximizes the 90th percentile DPS.
            for percentile in optimalRandomGearSet:
                if percentile in PercentileToOpt:
                    print("Optimizing " + percentile + "th percentile BiS")
                    limit = optimalRandomGearSet[percentile][1].getMateriaLimit()
                    curBest = optimalRandomGearSet[percentile][1]
                    counter = 0
                    while True:
                        print("Progress : " + str(counter)+"/"+str(limit))
                        if counter + depth > limit:
                            d = limit - counter
                        else:
                            d = depth
                        curBest, curMax, curRandom = materiaBisSolverV3(curBest, matGen, MateriaSpace, d, Fight, Fight.PlayerList[PlayerIndex].JobMod, IsTank, IsCaster, PlayerIndex,percentile,randomIteration,mendSpellSpeed, maxSPDValue=maxSPDValue)
                        counter += depth
                        if counter >= limit : break
                    optimalRandomGearSetMateria[percentile][0] = curMax
                    optimalRandomGearSetMateria[percentile][1] = deepcopy(curBest)
                    optimalRandomGearSetMateria[percentile][2] = deepcopy(curRandom)

            
    text = "Solver result " + ("(Using BF up-down)" if useNewAlgo else "(Using BF down-up)") + "\n"

    GearStat = optimalGearSet.GetGearSetStat(IsTank=IsTank)
    f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
    logging.getLogger("ffxivcalc").setLevel(level=logging.DEBUG)
    gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
    curMax, curRandom, duration, potency = preBakedFightGCDTierList[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration, getInfo=True)
    if "exp" in PercentileToOpt:
        statDict = optimalGearSet.GetGearSetStat(IsTank=IsTank)
        damageValue = computeDamageValue(statDict, JobMod, IsTank, IsCaster)
        gcdTimer = computeGCDTimer(statDict["SS" if IsCaster else "SkS"], hasteAmount)

        #optimalGearSet.addFood(curBestExpectedFood)
        text += "Best found gear set : "
        text += (str(optimalGearSet) + "\n")
        text += ("Expected Damage : " + str(curMax) + "\n")
        #text += ("Potency : " + str(potency))
        #text += ("TimeStamp : " + str(duration))
        #text += ("Random Damage : " + str(curRandom) + "\n")
        #text += str(computeDamageValue(optimalGearSet.GetGearSetStat(IsTank=IsTank), JobMod, IsTank, IsCaster))
        
        text += ("Base gcd timer (s) : " +  str(gcdTimer[0]) + "\n")
        if hasteAmount != 0 : text += ("Base gcd timer (s) : " +  str(gcdTimer[1]) + "\n")
        text += ("Determination damage increase : " + str(int(damageValue[1] * 100)/100) + "\n")
        if IsTank : text += ("Tenacity damage increase : " + str(int(damageValue[2] * 100)/100) + "\n")
        text += ("Speed DOT damage increase : " +  str(int(damageValue[3]*100)/100) + "\n")
        text += ("Crit rate : " + str(int(damageValue[4] * 100)/100) + "\n")
        text += ("Crit multiplier : " + str(int(damageValue[5] * 100)/100) + "\n")
        text += ("DH rate : " + str(int(damageValue[6] * 100)/100) + "\n")
        text += ("DH bonus (auto crit) : " + str(int(damageValue[7] * 100)/100) + "\n")
        text += "Weapon delay "+ str(weaponDelay) + " s"


    for percentile in optimalRandomGearSetMateria:
        GearStat = optimalRandomGearSetMateria[percentile][1].GetGearSetStat(IsTank=IsTank)
        f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
        gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
        optimalRandomGearSetMateria[percentile][0], optimalRandomGearSetMateria[percentile][2] = preBakedFightGCDTierList[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)
        text += (percentile + "th percentile gear :"+ "\n")
        text += (str(optimalRandomGearSetMateria[percentile][1])+ "\n")
        text += ("Expected Damage : " + str(optimalRandomGearSetMateria[percentile][0]) + "\n")
        text += ("Random Damage : " + str(optimalRandomGearSetMateria[percentile][2]) + "\n")
        text += str(computeDamageValue(optimalRandomGearSetMateria[percentile][1].GetGearSetStat(IsTank=IsTank), JobMod, IsTank, IsCaster))
        text += ("========================\n")

                             # Will find appropriate name for the file so it doesn't overwrite another existing file
    if saveAsFile : 
        filenameSkeleton = 'bisSolver' + JobEnum.name_for_id(Fight.PlayerList[PlayerIndex].JobEnum) + "Result"
        testFileName = filenameSkeleton
        counter = 1
        while os.path.isfile(testFileName + ".txt"):
            testFileName = filenameSkeleton + "(" + str(counter) + ")"
            counter += 1
        with open(testFileName + ".txt", 'w') as f:
            f.write(text)

    if not returnGearSet :
        return curMax, text

    return optimalGearSet, optimalRandomGearSetMateria, text
    
def materiaBisSolverV3(Set : GearSet, matGen : MateriaGenerator, matSpace : list[int], gcdTimerTierFight, hasteAmount : int,  JobMod : int, IsTank : bool, IsCaster : bool,PlayerIndex : int, 
                       percentile : str, randomIteration : int, mendSpellSpeed : bool,minSPDValue : int = 0, maxSPDValue : int = 5000, oversaturationIterationsPostGear : int = 0, 
                       findOptMateriaGearBF : bool = False, swapDHDetBeforeSpeed : bool = False, minPiety : int = 390):   
    """
    This function finds the best melds for the given Gear Set.

    The algorithm will first remove materias until the gear set is valid by removing the materias that are deemed
    less contributing. After that SpS/SkS will replace lowest impact materias until Speed values are achieved.

    THIS FUNCTION HAS NOT YET BEEN SHOWED TO BE OPTIMAL, BUT IT IS THE BEST ONE I HAVE FOUND.

    Set : GearSet -> GearSet to optimize
    matGen : MateriaGenerator -> Materia Generator
    matSpace : list[int] -> list of Materias to consider when oversaturating and removing.
    gcdTimerTierFight -> Dictionnary of different GCD timer and the given Fight object.
    hasteAmount : int -> possible haste a player can received. Affects GCD timer.
    JobMod : int -> Value of the JobMod of the player the Gear Set is on.
    IsTank : bool -> True if the player is a tank
    IsCaster : bool -> True if the player is a caster
    PlayerIndex : int -> Index of the player in the Fight's PlayerList
    percentile : str -> Percentile to optimize. "exp" is Expected.
    randomIteration : int -> Number of time to run random simulations.
    mendSpellSpeed : bool -> If true the algorithm will add SpS materias and not SkS
    maxSPDValue : int -> Maximum Speed value for the Gear Set with melds.
    minSPDValue : int -> Minimal Speed value for the Gear Set with melds.
    oversaturationIterationsPostGear : int -> Number of times the algorithm will oversaturate the gear set.
    findOptMateriaGearBF : bool -> If true means we are solving materias for every possible gear set/food. So this simply mutes the ProgressBar usually present.
    swapDHDetBeforeSpeed : bool -> If True, the solver will swap DH and Det before swapping melds with Speed materias. If False it swaps after.
    minPiety : int -> Minimum Piety value for the gear set. Default to 390
    """
    optimalSet = deepcopy(Set)
                                 # This first loop will forceAddMateria to the whole set until no more can be
    for gear in optimalSet:
        if gear.getIgnoreOptimize() : continue
        for type in matSpace:
            for j in range(oversaturationIterationsPostGear):
                if type != 3: 
                    for i in range(gear.getNumberPossibleMateria(type, matGen)) : gear.forceAddMateria(matGen.GenerateMateria(type))

                             # Will remove materias until the gearset is valid.
    if not findOptMateriaGearBF : 
        pB = ProgressBar.init(optimalSet.getNumberMateria() - optimalSet.getMateriaLimit(), "Removing oversaturated")

    while not optimalSet.hasValidMelding():

        curMaxDPS = 0
        curTypeToRemove = None
        
        for type in optimalSet.getMateriaTypeList(ignoreValidMeld=True):
                             # Will trial which materia type is the best to remove.
                             # SkS/SpS are not yet on the gear set, so they are ignored.
            trialSet = deepcopy(optimalSet)
            gearName = trialSet.removeFirstFoundMateriaInvalidPiece(type)

            GearStat = trialSet.GetGearSetStat(IsTank=IsTank)

            f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
            gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
            ExpectedDamage, randomDamageDict = gcdTimerTierFight[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)
            
            solver_logging.warning("Trial by removing " + StatType.name_for_id(type) + " from " + gearName + " : " + (str(ExpectedDamage) if percentile == "exp" else str(randomDamageDict[percentile])) + (" Expected : " + str(ExpectedDamage) if percentile != "exp" else ""))
            solver_logging.warning("Dict Stat : " + str(GearStat))
            if percentile == "exp" and ExpectedDamage > curMaxDPS:
                curMaxDPS = ExpectedDamage
                curTypeToRemove = type
            elif percentile != "exp" and randomDamageDict[percentile] > curMaxDPS:
                curMaxDPS = randomDamageDict[percentile]
                curTypeToRemove = type

                             # Will now remove the best meld to remove from oversaturated gear piece
        if optimalSet.removeFirstFoundMateriaInvalidPiece(curTypeToRemove) == None:
            solver_logging.error("Trying to remove materia that isn't illegal")
            raise InvalidMateriaSpace
        
        solver_logging.warning("Removing " + StatType.name_for_id(curTypeToRemove))
        if not findOptMateriaGearBF : next(pB)


    if swapDHDetBeforeSpeed:
        optimalSet, curMaxDPS = materiaDHAndDetSolver(curMaxDPS,optimalSet, matGen, hasteAmount, IsTank=IsTank, IsCaster=IsCaster, JobMod=JobMod, gcdTimerTierFight=gcdTimerTierFight,PlayerIndex=PlayerIndex, randomIteration=randomIteration)

    solver_logging.warning("Replacing materias until SpS/SkS values are achieved.")

    optimalSpeedSet = deepcopy(optimalSet)
    curMaxSpeedDPS = curMaxDPS

    GearStat = optimalSet.GetGearSetStat(IsTank=IsTank)
    pBTotal = max(0,int((maxSPDValue - GearStat["SS"])/matGen.EvenValue))

    haveFoundMinSPDSet = GearStat[("SS" if mendSpellSpeed else "SkS")] >= minSPDValue
    
    if not findOptMateriaGearBF : 
        pBTotal = max(1,int((maxSPDValue - GearStat["SS"])/matGen.EvenValue))
        pbReplace = ProgressBar.init(pBTotal, "Replacing by SpS/SkS")

    while (GearStat["SS"] if mendSpellSpeed else GearStat["SkS"]) + matGen.EvenValue < maxSPDValue:
        curMaxDPS = 0
        curTypeToReplace = None
        curGearPieceToReplace = None
        mat = matGen.GenerateMateria(StatType.SS if mendSpellSpeed else StatType.SkS)

        for type in optimalSet.getMateriaTypeList():
            trialSet = deepcopy(optimalSet)
            for gear in trialSet:
                if gear.getIgnoreOptimize(): continue
                             # Will look for first piece of gear that
                             # can have the desired type replaced by SpS or SkS
                if gear.hasStatMeld(type) and gear.canReplaceMateriaNoLoss(mat):
                    gear.removeMateriaType(type)
                    gear.AddMateria(mat)

                    GearStat = trialSet.GetGearSetStat(IsTank=IsTank)

                    f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                    gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
                    ExpectedDamage, randomDamageDict = gcdTimerTierFight[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)

                    solver_logging.warning("Trial be replacing " + StatType.name_for_id(type) + " from " + gear.getGearTypeName() + " : " + (str(ExpectedDamage) if percentile == "exp" else str(randomDamageDict[percentile])) + (" Expected : " + str(ExpectedDamage) if percentile != "exp" else ""))
                    
                    if percentile == "exp" and ExpectedDamage > curMaxDPS:
                        curMaxDPS = ExpectedDamage
                        curTypeToReplace = type
                        curGearPieceToReplace = gear.getGearTypeName()
                    elif percentile != "exp" and randomDamageDict[percentile] > curMaxDPS:
                        curMaxDPS = randomDamageDict[percentile]
                        curTypeToReplace = type
                        curGearPieceToReplace = gear.getGearTypeName()
                    break
        
                             # We will now replace the highest valued DPS materia with SpS or SkS
        if curTypeToReplace == None:
                             # If curTypeToReplace is still none, then no possible materias can be removed so we exit.
            solver_logging.warning("Cannot find any materia to replace")
            break
        
        optimalSet.removeMateriaSpecGear(curGearPieceToReplace, curTypeToReplace)
        optimalSet.GearSet[curGearPieceToReplace].AddMateria(mat)
        GearStat = optimalSet.GetGearSetStat(IsTank=IsTank)
        solver_logging.warning("Replacing " + StatType.name_for_id(curTypeToReplace) + " from " + curGearPieceToReplace + " Speed is now : " + str(GearStat[("SS" if mendSpellSpeed else "SkS")]))
        if not findOptMateriaGearBF : next(pbReplace)

                             # Will now compare with previous best to see if new Speed values are better
                             # Or take the first set has minSPDValue
        if curMaxDPS > curMaxSpeedDPS or ((not haveFoundMinSPDSet) and GearStat[("SS" if mendSpellSpeed else "SkS")] > minSPDValue):
            if not haveFoundMinSPDSet and GearStat[("SS" if mendSpellSpeed else "SkS")] > minSPDValue : 
                haveFoundMinSPDSet = True
                solver_logging.warning("Found minSPDValue set with SPD value : " + str(GearStat[("SS" if mendSpellSpeed else "SkS")]))
            curMaxSpeedDPS = curMaxDPS
            optimalSpeedSet = deepcopy(optimalSet)
            solver_logging.warning("Found new optimal Speed set with damage " + str(curMaxDPS) + " Speed value : " + str(GearStat[("SS" if mendSpellSpeed else "SkS")]))

    optimalSet = deepcopy(optimalSpeedSet)

    if not swapDHDetBeforeSpeed:
        optimalSet, curMaxDPS= materiaDHAndDetSolver(curMaxDPS,optimalSet, matGen, hasteAmount, IsTank=IsTank, IsCaster=IsCaster, JobMod=JobMod, gcdTimerTierFight=gcdTimerTierFight,PlayerIndex=PlayerIndex, randomIteration=randomIteration
                                                      )
    if minPiety > 390:
                             # Adding piety until min requirement are met.
        optimalSet = pietySolver(minPiety, curMaxDPS, optimalSet, matGen, hasteAmount, IsTank, IsCaster, 
                                JobMod, gcdTimerTierFight, PlayerIndex, randomIteration)
    return optimalSet, 0, {}
                
def materiaDHAndDetSolver(curMaxDPS : float, Set : GearSet, matGen : MateriaGenerator, hasteAmount : int, IsTank : bool, IsCaster : bool, 
                          JobMod, gcdTimerTierFight, PlayerIndex : int, randomIteration : int):
    """
    This function swapes DH and Det melds to see if an improvement to DPS can be made.

    Args:
        curMaxDPS (float): the DPS of the Set before the swaps are made
        Set (GearSet): Set to swap melds in
        matGen (MateriaGenerator): Materia Generator
        IsTank (bool): True if player is tank
        IsCaster (bool): True if player is caster
        JobMod (_type_): JobMod of the player
        gcdTimerTierFight (dict): Dictionnary of different GCD timer and the given Fight object.
        PlayerIndex (int): Index of the player in the Fight's PlayerList
        randomIteration (int): number of random iterations.

    Returns:
        Optimized gear set
    """
    solver_logging.warning("Exploring DH/Det replacement")

    optimalSet = deepcopy(Set)

    trialSetDH = deepcopy(optimalSet)
    trialSetDHCurMaxDPS = 0
    trialSetDet = deepcopy(optimalSet)
    trialSetDetCurMaxDPS = 0

    while True:

        hasChangedMeld = False

        for gear in trialSetDH:
            if gear.getIgnoreOptimize(): continue
            if gear.hasStatMeld(StatType.Det) and gear.canReplaceMateriaNoLoss(matGen.GenerateMateria(StatType.DH)):
                hasChangedMeld = True
                gear.removeMateriaType(StatType.Det)
                gear.AddMateria(matGen.GenerateMateria(StatType.DH))

                GearStat = trialSetDH.GetGearSetStat(IsTank=IsTank)

                f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
                ExpectedDamage, randomDamageDict = gcdTimerTierFight[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)

                if ExpectedDamage > curMaxDPS and ExpectedDamage > trialSetDHCurMaxDPS :
                    solver_logging.warning("Found Better DH : " + str(GearStat))
                    optimalSet = deepcopy(trialSetDH)
                    trialSetDHCurMaxDPS = ExpectedDamage

        if not hasChangedMeld : break

    while True:

        hasChangedMeld = False

        for gear in trialSetDet:
            if gear.getIgnoreOptimize(): continue
            if gear.hasStatMeld(StatType.DH) and gear.canReplaceMateriaNoLoss(matGen.GenerateMateria(StatType.Det)):
                hasChangedMeld = True
                gear.removeMateriaType(StatType.DH)
                gear.AddMateria(matGen.GenerateMateria(StatType.Det))

                GearStat = trialSetDH.GetGearSetStat(IsTank=IsTank)

                f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
                ExpectedDamage, randomDamageDict = gcdTimerTierFight[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)
                
                if ExpectedDamage > curMaxDPS and ExpectedDamage > trialSetDHCurMaxDPS and ExpectedDamage > trialSetDetCurMaxDPS :
                    solver_logging.warning("Found Better Det : " + str(GearStat))
                    optimalSet = deepcopy(trialSetDH)
                    trialSetDetCurMaxDPS = ExpectedDamage

        if not hasChangedMeld : break

    return optimalSet, max(trialSetDHCurMaxDPS, trialSetDetCurMaxDPS)

def pietySolver(minPiety : int, curMaxDPS : float, Set : GearSet, matGen : MateriaGenerator, hasteAmount : int, IsTank : bool, IsCaster : bool, 
                JobMod, gcdTimerTierFight, PlayerIndex : int, randomIteration : int):
    """
    This works the same way as replacing SkS/SpS materia for min Speed value. It will look for the materia that when removed results
    in the lowest DPS loss and will change this one for a Piety Materia.
    """

    solver_logging.warning("Replacing materias with Piety to meet minimum piety requirement.")

    optimalSet = deepcopy(Set)
    GearStat = optimalSet.GetGearSetStat(IsTank=IsTank)
    mat = matGen.GenerateMateria(StatType.Piety)

    while GearStat["Piety"] < minPiety:

        trialBestDPS = 0
        curTypeToReplace = None
        curGearPieceToReplace = None

        for type in optimalSet.getMateriaTypeList():

            trialSet = deepcopy(optimalSet)

                             # Will now look for the least DPS loss materia to replace.
            for gear in trialSet:
                if gear.getIgnoreOptimize(): continue
                if gear.hasStatMeld(type) and gear.canReplaceMateriaNoLoss(mat):
                    gear.removeMateriaType(type)
                    gear.AddMateria(mat)

                    GearStat = trialSet.GetGearSetStat(IsTank=IsTank)

                    f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                    gcdTimer = computeGCDTimer(GearStat["SS" if IsCaster else "SkS"],hasteAmount)
                    ExpectedDamage, randomDamageDict = gcdTimerTierFight[gcdTimer].SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)

                    solver_logging.warning("Trial be replacing " + StatType.name_for_id(type) + " from " + gear.getGearTypeName() + " : " + (str(ExpectedDamage)) + (" Expected : " + str(ExpectedDamage)))

                    if ExpectedDamage > trialBestDPS:
                        trialBestDPS = ExpectedDamage
                        curTypeToReplace = type
                        curGearPieceToReplace = gear.getGearTypeName()

        if curTypeToReplace == None or curGearPieceToReplace == None:
            solver_logging.warning("Could not find another materia to replace with Piety")
            return optimalSet

                            # Replacing found materia to replace with Piety
        try:
            optimalSet.removeMateriaSpecGear(curGearPieceToReplace, curTypeToReplace)
            optimalSet.GearSet[curGearPieceToReplace].AddMateria(mat)
            GearStat = optimalSet.GetGearSetStat(IsTank=IsTank)
        except:
            input(str(curTypeToReplace))
            print(curGearPieceToReplace)

    
    return optimalSet






"""

OLD STUFF

def materiaBisSolver(Set : GearSet, matGen : MateriaGenerator, matSpace : list[int], maxDepth : int, Fight, JobMod : int, IsTank : bool, IsCaster : bool,PlayerIndex : int, percentile : str, randomIteration : int,mendSpellSpeed : bool,maxSPDValue : int = 5000):
    

    NOT RECOMMENDED TO USE SINCE SLOW AND DEPRECATED

    This functions solves the materia BiS for a given gearset.
    trialSet : GearSet -> GearSet for which to optimize Materias
    matGen : MateriaGenerator -> Materia Generator object
    matSpace : list -> list of all the stats the materia solver will use for melds.
    maxDepth : int -> Up to what depth to search
    Fight : Fight -> Fight instance
    JobMod : int -> JobMod of the player
    IsTank : bool -> If is a tank
    percentile : str -> percentile of the optimization. exp is expected
    maxSPDValue : int -> max Spell Speed or Skill Speed value. Any gear set with higher than that value will be discarded.
    
    GearSetDict = {
        "99" : [0,GearSet()],
        "90" : [0,GearSet()],
        "75" : [0,GearSet()],
        "50" : [0,GearSet()],
        "exp" : [0,GearSet()]
    }
    curOptimalDPS = 0

    result = []
    matRange = len(matSpace)
    pB = ProgressBar.init(len(matSpace)**maxDepth, "Finding best " + str(maxDepth) + " next meld")
    def solver(curDepth, matList, curOptimalDPS):
        for i in range(matRange):
            if curDepth > 1 :
                             # Going recursively until at the max depth
                solver(curDepth - 1, matList, curOptimalDPS)
            else:
                             # Will try to put all the materias in the matList onto the gear pieces.
                trialSet = deepcopy(Set)
                goNext = False
                for matType in matList:
                    newMateria = matGen.GenerateMateria(matSpace[matType])
                
                             # Will find the first piece of gear on which the materia can go
                    key = trialSet.findFirstPieceMateria(newMateria)
                             # If none we ignore this gear set. for now
                    if key == None:
                        goNext = True
                    else:
                        trialSet.GearSet[key].AddMateria(newMateria)
                if not goNext: 
                    GearStat = trialSet.GetGearSetStat(IsTank=IsTank)
                    if not ((mendSpellSpeed and GearStat["SS"] > maxSPDValue) or (not mendSpellSpeed and GearStat["SkS"] > maxSPDValue)):
                        f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto = computeDamageValue(GearStat, JobMod, IsTank, IsCaster)
                        ExpectedDamage, randomDamageDict = Fight.SimulatePreBakedFight(PlayerIndex, GearStat["MainStat"],f_WD, f_DET, f_TEN, f_SPD, f_CritRate, f_CritMult, f_DH, DHAuto, n=randomIteration)

                        result.append([ExpectedDamage, deepcopy(trialSet), randomDamageDict])
                next(pB)
            matList[curDepth - 1] += 1
        matList[curDepth - 1] = 0

    matList = [0 for i in range(maxDepth)]
                             # solver will find all possible arrangements of materias of the given matSpace and try all of them.
    solver(maxDepth, matList, curOptimalDPS)
                             # result now has all the result. So we take the max of each percentile.
                             # The metric used here the dps corresponding to the percentile or expected.
    curMaxDPS = 0
    curExpected = 0
    curBestSet = None
    curBestRandom = {}
    if percentile == "exp":
        for gSet in result:
            if gSet[0] > curMaxDPS:
                curMaxDPS = gSet[0]
                curExpected = curMaxDPS
                curBestSet = gSet[1]
                curBestRandom = gSet[2]
    else:
        for gSet in result:
            if gSet[2][percentile] > curMaxDPS:
                curMaxDPS = gSet[2][percentile]
                curExpected = gSet[0]
                curBestSet = gSet[1]
                curBestRandom = gSet[2]

    return curBestSet, curExpected, curBestRandom
"""



        




    




    






"""
This code was written by Apollo (apollo.van.waddleburg on discord) and adapted to fit the needs of the sim.
"""

"""Defines utilities for determining important prepull buffs from fflogs."""
import logging
from typing import List

import pandas as pd

from ffxivcalc.Request import custom_columns

LOG = logging.getLogger(__name__)

# TODO: export into a configurable and accessable format
ALL = {
    "Medicated": 30,
    "Swiftcast": 10,
}

WAR = {
    "Inner Release": 15,
}

PLD = {
    "Fight or Flight": 20,
}

DRK = {
    "Delirium": 15,
    "Blood Weapon": 15,
    "Blackest Night": 7,
}

GNB = {
    "No Mercy": 20,
}

RPR = {
    "Arcane Circle": 20,
}

MNK = {
    "Riddle of Fire": 20,
    "Riddle of Wind": 15,
    "Brotherhood": 15,
}

DRG = {
    "Life Surge": 5,
    "Lance Charge": 20,
    "Battle Litany": 15,
    "Right Eye": 20,
    "Left Eye": 20,
}

NIN = {
    "Kassatsu": 15,
}

SAM = {
    "Meikyo Shisui": 15,
}

BRD = {
    "Battle Voice": 15,
}

DNC = {
    "Standard Finish": 30,
    "Technical Finish": 20,
    "Devilment": 20,
    # "Flourish":
}

MCH = {}

RDM = {
    "Acceleration": 20,
    "Dualcast": 5,
    "Embolden": 20,
}

BLM = {
    "Triplecast": 10,
    "Sharpcast": 30,
    "Ley Lines": 30,
}

SMN = {
    "Searing Light": 30,
}

WHM = {
    "Presence of Mind": 15,
}

AST = {
    "the Arrow": 15,
    "the Balance": 15,
    "the Ewer": 15,
    "the Spire": 15,
    "the Spear": 15,
    "the Bole": 15,
    "Divination": 15,
    "Earthly Dominance": 10,
    "Giant Dominance": 20,
}

SCH = {}

SGE = {
    "Eukrasia": 60,  # has no duration, 30 is fine.
}


POTENTIAL_PREPULL_BUFFS = {
    **ALL,
    **AST,
    **BLM,
    **BRD,
    **DNC,
    **DRG,
    **DRK,
    **GNB,
    **MCH,
    **MNK,
    **NIN,
    **RPR,
    **PLD,
    **SAM,
    **SGE,
    **SCH,
    **SMN,
    **WAR,
    **WHM,
}


def prepull_aura_frame(fight_events: pd.DataFrame) -> pd.DataFrame:
    """Builds a frame depicting prepull auras."""
    output = []
    combatantinfo = fight_events[
        fight_events[custom_columns.TYPE] == "combatantinfo"
    ]
    for (
        fight_time,
        source_id,
        source_name,
        source_type,
        source_subtype,
        auras,
    ) in zip(
        combatantinfo[custom_columns.FIGHT_TIME],
        combatantinfo[custom_columns.SOURCE_ID],
        combatantinfo[custom_columns.SOURCE_NAME],
        combatantinfo[custom_columns.SOURCE_TYPE],
        combatantinfo[custom_columns.SOURCE_SUBTYPE],
        combatantinfo["auras"],
    ):
        for aura in auras:
            output.append(
                {
                    custom_columns.FIGHT_TIME: fight_time,
                    custom_columns.SOURCE_ID: source_id,
                    custom_columns.SOURCE_NAME: source_name,
                    custom_columns.SOURCE_TYPE: source_type,
                    custom_columns.SOURCE_SUBTYPE: source_subtype,
                    **aura,
                }
            )

    return pd.DataFrame(output)


def infer_prepull_buff_casts(fight_events: pd.DataFrame) -> pd.DataFrame:
    """Infers all prepull buffs applications into a DF with all relevant data.

    Specifically, does the following:
        - Looks at the combatant info row to get starting buffs
        - For each one, looks for when the buff falls off in combat
        - From this, inferences the original application time
    """
    # Get the specific subframe for particular events.
    outputs: List[pd.DataFrame] = []

    # Get a quick subframe for the first refresh or removal of a buff
    buff_removal_or_refresh = (
        fight_events[
            (
                fight_events[custom_columns.TYPE].isin(
                    [custom_columns.REMOVE_BUFF, "refreshbuff"]
                )
            )
        ]
        .sort_values(by=[custom_columns.FIGHT_TIME], axis=0)
        .dropna(how="all", axis=1)
        .reset_index(drop=True)
    )

    # groupby each player, build outputs per player
    for player_name, info_data in fight_events[
        fight_events[custom_columns.TYPE] == "combatantinfo"
    ].groupby(by=custom_columns.SOURCE_NAME):
        LOG.debug(f"Finding auras for {player_name}")
        # while info data should be a single row, just loop through it for safety.
        for auras in info_data["auras"]:
            for aura in auras:
                # Check aura-names
                duration = POTENTIAL_PREPULL_BUFFS.get(aura["name"])
                if duration is None:
                    LOG.debug(f"Skipping {aura}, not tracking.")
                    continue
                LOG.debug(f"Processing auras {aura}.")

                # get the earliest removal or refresh of the buff
                specific_buff_removal_or_refresh = buff_removal_or_refresh[
                    (
                        buff_removal_or_refresh[custom_columns.TARGET_NAME]
                        == player_name
                    )
                    & (
                        buff_removal_or_refresh[custom_columns.ABILITY_NAME]
                        == aura["name"]
                    )
                ]

                if specific_buff_removal_or_refresh.empty:
                    LOG.warning(f"Unable to find removal event for {aura}")
                    continue

                earliest_buff_reference = specific_buff_removal_or_refresh.head(
                    1
                ).copy(deep=True)
                # Inference the original usage, via subtracting the fight-time
                earliest_buff_reference.iloc[
                    0,
                    earliest_buff_reference.columns.get_loc(
                        custom_columns.FIGHT_TIME
                    ),
                ] -= duration
                earliest_buff_reference.iloc[
                    0,
                    earliest_buff_reference.columns.get_loc(
                        custom_columns.TYPE
                    ),
                ] = custom_columns.ESTIMATED_APPLY_BUFF
                outputs.append(earliest_buff_reference)

    if not outputs:
        return pd.DataFrame()

    return (
        pd.concat(outputs, axis=0)
        .round({custom_columns.FIGHT_TIME: 3})
        .sort_values(by=[custom_columns.FIGHT_TIME], axis=0)
        .dropna(how="all", axis=1)
        .reset_index(drop=True)
    )


"""CLI entrypoint for FFXIV-Combat-Simulator"""
from argparse import ArgumentParser
from pathlib import Path
from sys import exit
import logging

from ffxivcalc import __name__ as prog
from ffxivcalc.UI.SimulationInput import ExecuteMemoryCode as fight_main
from ffxivcalc.Tester.testImplementation import executeTests
from ffxivcalc.UI.TUI import TUI_draw

__logger__ = logging.getLogger("ffxivcalc") # root logger

def get_parser() -> ArgumentParser:
    """Defines all the cli arguments to be parsed
    :return: An ArgumentParser object
    """
    verbose_help = " -v (Displays on terminal logging at level ERROR), -vv (Displays on terminal logging at level WARNING), -vvv (Creates a log file that has logging level DEBUG)"
    parser = ArgumentParser(prog=prog)
    parser.add_argument('-v', '--verbose', action='count', default=0, help=verbose_help)
    subparsers = parser.add_subparsers(help='action to perform', dest='action')

    # Running in code simulation
    simulation_parser = subparsers.add_parser('simulate')
    simulation_parser.add_argument('-code_simulation', '--code-simulation', type=bool, default=False)
    simulation_parser.add_argument('-s', '--step-size', type=float, default=0.01)
    simulation_parser.add_argument('-l', '--time-limit', type=int, default=1000)

    # Running tester
    test_parser = subparsers.add_parser('tests')
    seed_help = "-s XXXXXX (sets the given seed for the execution of the tests. Relevant to DOT and AA test suite. Sets the seed for both the DOT and AA test suite.)"
    test_parser.add_argument('-s', '--seed', type=int, default=0, help=seed_help) # argument to specify seed for DOT and aa test suite.
                                                                  # The same seed is given to both test if specified.
    name_help = "-n XXXXX (Name of a specific test suite to run. If not specified runs all the test suites. If invalid name returns an error).\nA valid name is the 3 letters of the test suite's name."
    test_parser.add_argument('-n', '--name', type=str, default="", help=name_help) # Specifies te name of the test suites we want to run. 

    # Running tui
    subparsers.add_parser('tui')

    return parser


def main() -> int:
    """
    Entrypoint into the cli. Can be called with ``python -m ffxivcalc``
    :return: Return code, for use with cli programs to indicate success/failure
    """
    parser = get_parser()
    args = parser.parse_args()
    level = logging.ERROR

    if args.verbose > 0:
        match args.verbose:
            case 1: # displays error and above
                level = logging.ERROR
                logging.basicConfig(format='[%(levelname)s] %(name)s : %(message)s',level=level)
            case 2: # displays warning and above
                level = logging.WARNING
                logging.basicConfig(format='[%(levelname)s] %(name)s : %(message)s',level=level)
            case 3: # saves everything to a save file
                level = logging.DEBUG
                logging.basicConfig(format='[%(levelname)s] %(name)s : %(message)s',filename='ffxivcalc_log.log', encoding='utf-8',level=level)
            case _: # more than 3 V is too much, save as case 3
                level = logging.DEBUG
                logging.basicConfig(format='[%(levelname)s] %(name)s : %(message)s',filename='ffxivcalc_log.log', encoding='utf-8',level=level)
        __logger__.setLevel(level=level) 

    match args.action:
        case 'simulate':
            fight_main(False, time_unit=args.step_size, TimeLimit=args.time_limit)
        case 'tests':
            executeTests(setSeed=args.seed, testSuiteName=args.name,level=level)
        case 'tui':
            TUI_draw()
        case _: # default case
            parser.print_help()

    return 0


if __name__ == '__main__':
    exit(main())


"""
This file contains the class for representing gear pieces. It will also contain the class
GearSet which represents a gear set. A player can be given a gear set and will see its stats change
according to the gear set. Every piece of gear will act the same way a piece does in the game.
It will have possible melds (and a limit) and will have stat caps such that some melds do nothing.

This will also contain the Enum "GearType" which will differentiate the type of piece a gear is,
"Food" which can affect a player's stats, Materias which will affect a gear's stats.
"""

from ffxivcalc.helperCode.exceptions import MateriaOverflow, InvalidStatRequest, InvalidFileName
import json

from enum import IntEnum

class GearType(IntEnum):
    WEAPON = 0
    SHIELD = 1               # Only Paladin has a shield
    HEAD = 2
    BODY = 3
    HANDS = 4
    LEGS = 5
    FEET = 6
    EARRINGS = 7
    NECKLACE = 8
    BRACELETS = 9
    LRING = 10               # Left Ring
    RING = 11                # Right Ring

    @classmethod
    def name_for_id(cls, id : int) -> str:
        # maps from id -> name
        if id in cls.__members__.values():
            return cls(id).name
        return 'INVALID'      
    
    @classmethod
    def id_for_name(cls, name : str) -> int:
        # maps from name -> id
        if name in cls.__members__.keys():
            return cls[name].value
        return -1 # Evaluated as Unknown

class StatType(IntEnum):
    Crit = 0
    DH = 1
    Det = 2
    SS = 3
    SkS = 4
    Ten = 5
    Piety = 6
    MainStat = 7
    WD = 8      

    @classmethod
    def name_for_id(cls, id : int) -> str:
        # maps from id -> name
        if id in cls.__members__.values():
            return cls(id).name
        return 'Unknown'      
    
    @classmethod
    def id_for_name(cls, name : str) -> int:
        # maps from name -> id
        if name in cls.__members__.keys():
            return cls[name].value
        return -1 # Evaluated as Unknown

class Stat:
    """
    This class represents a stat on a gear piece.
    StatType -> Type of the stat
    Value -> Value of the stat
    """

    def __init__(self, StatType : StatType, Value : int):
        self.StatType = StatType
        self.Value = Value

    def __str__(self):
        return StatType.name_for_id(self.StatType) + " : " + str(self.Value)
    
class Food:
    """
    This class represents food. Food gives a stat bonus that caps at a certain value.
    A food can be "attached" to a GearSet object and will buff it.
    statBonusDict : dict -> Dictionnary where key is the StatType's Name and the value is the cap of that stat with the percentage bonus.
    name : str -> Name of the food
    """

    def __init__(self, statBonusDict : dict, name : str):
        self.name = name
        self.statBonusDict = statBonusDict

    def __str__(self):
        strGenerator = (key + " " + str(int( 100 * self.statBonusDict[key][1])) + "% max of " + str(self.statBonusDict[key][0]) + " |" for key in self.statBonusDict)
        returnStr = "Food " + self.name + " gives -> "
        for gen in strGenerator:
            returnStr += gen
        return returnStr

    def getLimitStatBonus(self, StatName : str) -> int:
        """
        This function returns the limit of a stat for the food.
        StatName : str -> Name of the StatType.
        """

        return self.statBonusDict[StatName][0] if StatName in self.statBonusDict.keys() else 0
    
    def getPercentStatBonus(self, StatName : str) -> float:
        """
        This function  returns the percent bonus from the food for the given StatName
        StatName : str -> Name of the StatType.
        """
        return self.statBonusDict[StatName][1] if StatName in self.statBonusDict.keys() else 0
    

class Materia:
    """
    This class represents a Materia. Gear pieces can have materias,
    Materias have certain types. The increment in stat value is created when creating a "MateriaGenerator"
    object that will generate materias when requested. This lets the user create arbitrary Materias.
    StatType -> Type of the stat
    Value : int -> Value of the bonus stat given by this materia
    """

    def __init__(self, StatType : StatType, Value : int):
        self.StatType = StatType
        self.Value = Value
                             # Information related to if the materia wastes values.
        self.wasteValue = False
        self.wasteAmount = 0

    def __str__(self):
        return StatType.name_for_id(self.StatType) + " : " + str(self.Value) + (" Waste : " + str(self.wasteAmount) if self.wasteValue else "")

class MateriaGenerator:
    """
    This class will generate Materias. It will allow to specify an Even numbered bonus and an odd numbered bonus
    OddValue : int -> Bonus stat given if materia is odd numbered
    EvenValue : int -> Bonus stat given if materia is even numbered
    """
    
    def __init__(self, OddValue : int, EvenValue : int):
        self.OddValue = OddValue
        self.EvenValue = EvenValue

    def GenerateMateria(self,StatType : StatType, Even : bool = True) -> Materia:
        """
        This function generates a Materia. Defaults to an even numbered materia
        StatType -> Type of the stat
        Even : bool -> True if even numbered. False if odd (Defaults to True)
        """
        return Materia(StatType, self.EvenValue if Even else self.OddValue)

class Gear:
    """
    This class represent a gear piece.
    GearType -> Type of the gear piece
    StatList : list(Stat) -> List of all stat of the gear piece.
    MateriaLimit : int -> Limit of materia the gear can receive.
    weaponDelay : int -> Weapon delay for AAs. Base value of 0. Can be set using setWeaponDelay and accessed using getWeaponDelay on the object
    Name : str -> Used to differentiate same type gear

    __ignoreOptimize : bool -> If true, solver will not attempt to optimize the materias on this gear piece.


    """
    def __init__(self, GearType : GearType, StatList : list, MateriaLimit : int = 2, Name : str = ""):
        self.GearType = GearType
        self.Stat = {}
        self.Name = Name
        self.StatLimit = 0   # StatLimit of a gear is always equal to the highest stat of it.
        self.illegalMeld = False # This is set to true when this piece of gear is illegaly Overmelded

        self.__ignoreOptimize = False 

        self.weaponDelay = 0 # Can be set using setWeaponDelay

        for Stats in StatList:
            self.Stat[StatType.name_for_id(Stats.StatType)] = Stats
                             # We look for the highest stat value
            if Stats.StatType != StatType.MainStat and Stats.Value > self.StatLimit : self.StatLimit = Stats.Value

        self.Materias = []   # List of Materia object associated
        self.MateriaLimit = MateriaLimit
        self.MateriasCount = 0

    def __str__(self):
        strGenerator = (str(self.Stat[key]) + " | " for key in self.Stat)
        materiaGenerator = (str(mat) + " | " for mat in self.Materias) if self.MateriasCount > 0 else ("" for i in range(0))
        strReturn = GearType.name_for_id(self.GearType) + " : "
        for stat in strGenerator:
            strReturn += stat
        strReturn += " Materias : "
        for mat in materiaGenerator:
            strReturn += mat

        return strReturn + " Name : " + self.Name
    
    def getWeaponDelay(self):
        """Return value of weapon delay. If not a weapon returns 0
        """
        return self.weaponDelay
    
    def setWeaponDelay(self, value : float):
        """This function sets the value of weaponDelay for this gear
        """
        self.weaponDelay = value
    
    def setIgnoreOptimize(self, newVal : bool) -> None:
        """
        This function sets the value of self.__ignoreOptimize to the given newVal.
        newVal : bool -> New value to set the field too.
        """
        self.__ignoreOptimize = newVal

    def getIgnoreOptimize(self) -> bool:
        """
        This returns the value of the gear for self.__ignoreOptimize
        """
        return self.__ignoreOptimize

    def __addMateria(self, newMateria : Materia):
        """
        Private function called to add materia to the gear. This function is called by AddMateria and forceAddMateria.
        """
        StatName = StatType.name_for_id(newMateria.StatType)
        newStatValue = self.GetStat(StatName=StatName) + newMateria.Value
        if newStatValue > self.StatLimit:
                             # If true means the materia is wasting ressources.
            newMateria.wasteAmount = newStatValue - self.StatLimit
            newMateria.wasteValue = True

        self.Materias.append(newMateria)
        self.MateriasCount += 1

    def __hasStatType(self, type : StatType) -> bool:
        """
        This private function returns weither the given StatType has non zero value for this gear.
        """
        return (StatType.name_for_id(type) in self.Stat.keys()) and self.hasStatMeld(type)
    
    def setStatLimit(self, newLimit : int):
        """
        This function sets the gear stats' limit to the given newLimit value.

        Args:
            newLimit (int): new stat limit.
        """

        self.StatLimit = newLimit

    def AddMateria(self, newMateria : Materia):
        """
        This function adds a Materia Object to the gear. It will also flag the materia if stat is wasted.
        """

        if len(self.Materias) >= self.MateriaLimit :
            raise MateriaOverflow

        self.__addMateria(newMateria)

    def forceAddMateria(self, newMateria : Materia):
        """
        This function force adds a materia onto the gear. If the gear has reached its maximum materia count
        it will be flagged as "illegal". This function will never raise the MateriaOverflow error.
        """

        if len(self.Materias) >= self.MateriaLimit:
            self.illegalMeld = True

        self.__addMateria(newMateria)

    def hasValidMelding(self) -> bool:
        return not self.illegalMeld

    def getNumberPossibleMateria(self, type : StatType, matGen : MateriaGenerator) -> int:
        """
        This function returns the total number of materia that can be added to the gear with no loss
        of the given stat type. It uses the matGen to find the stat bonus from these materias. It is assumed
        we will use even numbered materia.
        type : StatType -> Type of the stat
        matGen : MateriaGenerator -> Materia generator used to add materias.
        """
        statName = StatType.name_for_id(type)
        statLimitMateria = (self.StatLimit - self.GetStat(StatName=statName)) // matGen.EvenValue
        return min(self.MateriaLimit,statLimitMateria)

    def getMateriaNumber(self, dictData : dict):
        """
        This function returns the number of materia of each StatTypes currently on this gear piece as well
        as if the materia is even or odd.
        (Out) dictData : dict -> This dictionnary is filled with the information. It can hence be passed on every gear pieces of the gear set.
        """

        for mat in self.Materias:
            statName = StatType.name_for_id(mat.StatType)
            if statName in dictData.keys():
                dictData[statName] += 1
            else :
                dictData[statName] = 1

    def canAddMateriaNoLoss(self, newMateria : Materia) -> bool:
        """
        This function returns weither the materia can be added to the gear piece
        without stat loss.
        """
        StatName = StatType.name_for_id(newMateria.StatType)
        return (self.MateriasCount < self.MateriaLimit and 
                (self.GetStat(StatName) if StatName in self.Stat.keys() else 0) + newMateria.Value <= self.StatLimit)
    
    def canReplaceMateriaNoLoss(self, newMateria : Materia) -> bool:
        """
        This function returns True if a materia of the gear piece can be replaced by a new materia
        without stat loss
        """
        StatName = StatType.name_for_id(newMateria.StatType)
        return (self.GetStat(StatName) if StatName in self.Stat.keys() else 0) + newMateria.Value <= self.StatLimit
    
    def hasMateriaType(self, type : StatType) -> bool:
        """
        This function returns weither the gear piece has a materia of the given type
        """
        for mat in self.Materias:
            if mat.StatType == type : return True
        return False

    def ResetMateriaSlot(self):
        """
        This function resets the materias of the gear
        """
        self.Materias = []
        self.MateriasCount = 0

    def removeMateriaType(self, statType : StatType):
        """
        This function removes the first materia of the given StatType found on the gear
        """
        for mat in self.Materias:
            if mat.StatType == statType:
                self.MateriasCount -= 1
                self.Materias.remove(mat)
                break
                        # Check if now is valid meld
        if self.illegalMeld:
            if self.MateriasCount <= self.MateriaLimit: self.illegalMeld = False

    def hasStatMeld(self, type : StatType) -> bool:
        """
        This function returns true if the gear has the given stattype melded to it.
        """

        for mat in self.Materias:
            if mat.StatType == type : return True
        return False

    def GetStat(self, StatName : str = "", StatEnum : int = -2) -> int:
        """
        This function returns the  statv alue associated with thhe given StatEnum value
        or the StatName.
        """

        if StatName == "" and StatEnum == -2 : raise InvalidStatRequest

        statValue = 0

        if StatEnum != -2 :
            name = StatType.id_for_name(StatEnum)
            statValue = self.Stat[name].Value if StatName in self.Stat.keys() else 0
        elif StatName != "":
            statValue = self.Stat[StatName].Value if StatName in self.Stat.keys() else 0

        StatEnum = StatType.id_for_name(StatName) if StatEnum == -2 else StatEnum

        for Materia in self.Materias:
            if Materia.StatType == StatEnum:
                statValue = min(Materia.Value + statValue, self.StatLimit)
        
        return int(statValue)
    
    def getGearTypeName(self) -> str:
        """
        This function returns the type of the gear.
        """
        return GearType.name_for_id(self.GearType)
    
    def getMateriaTypeList(self, IgnoreSpeedMateria : bool = True,IgnorePietyMateria : bool = True) -> list[StatType]:
        """
        This function returns a list of all present materia's stattype on the gear.
        """
        statTypeList = []
        for mat in self.Materias:
            if not (mat.StatType in statTypeList) and not ((mat.StatType == StatType.SS or mat.StatType == StatType.SkS) and IgnoreSpeedMateria) and not ((mat.StatType == StatType.Piety) and IgnorePietyMateria): 
                statTypeList.append(mat.StatType)
        return statTypeList
    
    def getNumberMateria(self) -> int:
        """
        This function returns the number of materia on this gear
        """
        return len(self.Materias)
    
class GearSet:
    """
    This class corresponds to a gearset. A player can have a gear set. A gear set is a list of Gear.
    """

    def __init__(self):
        self.GearSet = {}
        self.Food = None

    def __iter__(self):
        self.gearSetIter = iter(self.GearSet)
        return self

    def __next__(self):
        key = next(self.gearSetIter)
        return self.GearSet[key]

    def removeMateriaType(self, StatType):
        """
        This function removes the first materia of the valid StatType found in every gear piece.
        """
        for key in self.GearSet:
            self.GearSet[key].removeMateriaType(StatType)

    def removeFirstFoundMateriaInvalidPiece(self, type : StatType) -> str:
        """
        This function removes the first found materia of the given type on an invalid gear piece.
        If none is found the function returns None and otherwise it returns the name of the gear piece.
        """

        for gear in self:
            if not gear.hasValidMelding() and gear.hasStatMeld(type): 
                gear.removeMateriaType(type)
                return gear.getGearTypeName()
        return None

    def removeMateriaSpecGear(self, gearName : str, type : StatType):
        """
        This function removes the first materia of the given type from the given gear of the gear set
        """

        if gearName in self.GearSet.keys():
            self.GearSet[gearName].removeMateriaType(type)

    def addFood(self, newFood : Food):
        """
        This function adds newFood to the gearset.
        """
        self.Food = newFood

    def removeFood(self):
        """
        This function removes the food on the gear set.
        """
        self.Food = None

    def __str__(self):
        gearInfoGenerator = (str(self.GearSet[gear]) + "\n" for gear in self.GearSet)
        GearInfo = ""
        for info in gearInfoGenerator:
            GearInfo += info

        return "Gearset's info:\n" + GearInfo + " Final Stats : " + str(self.GetGearSetStat(IsTank=self.IsTank)) + "\n" + self.strMateriaNumber() + "\n" + str(self.Food)

    def findFirstPieceMateria(self, newMateria : Materia) -> str:
        """
        This function returns the key to the first piece of gear
        in the gear set that can have the newMateria attached to it
        without loss. If none is found, the function returns None
        """
        for key in self.GearSet:
            if self.GearSet[key].canAddMateriaNoLoss(newMateria):
                return key
        return None

    def AddGear(self, newGear : Gear):
        """
        This function is used to add a new gear to the gear set. If a gear of the type is already present it is switched.
        """
        self.GearSet[GearType.name_for_id(newGear.GearType)] = newGear

    def strMateriaNumber(self) -> str:
        """
        This function find the number of materia in the gear set and displays it in an easy way to understand.
        """

        dataDict = {}

        for key in self.GearSet:
            self.GearSet[key].getMateriaNumber(dataDict)
        
        returnStr = "Materia numbers : "

        for stat in dataDict:
            returnStr += stat + " " + str(dataDict[stat]) + "x" + " |"
        
        return returnStr

    def RemoveGear(self, Type : int):
        """
        This function removes the gear of the given type. If the gear set does not have the gear type then nothing happens.
        """
        name = GearType.name_for_id(Type)
        if name in self.GearSet:
            self.GearSet.pop(name)

    def getMateriaLimit(self) -> int:
        """
        This function returns the total number of materia the GearSet can receive, as well
        as the total of Odd and Even it can receive.

        FOR NOW ASSUMES ALL MATERIAS ARE EVEN
        """
        limit = 0
        for key in self.GearSet:
            limit += self.GearSet[key].MateriaLimit
        return limit
    
    def getMateriaTypeLimit(self, type : StatType, matGen : MateriaGenerator) -> int:
        """
        This function returns the total number of even materia (of the given MateriaGenerator object) of a given type the gear set can additionally receive.
        
        type : StatType -> Type of the materia
        matGen : MateriaGenerator : -> Materia Generator
        """

        limit = 0
        for gear in self:
            limit += gear.getNumberPossibleMateria(type, matGen)
        return limit
    
    def getNumberMateria(self) -> int:
        """
        This function returns the current number of materia on the gear set.
        """
        total = 0
        for gear in self: total += gear.getNumberMateria()
        return total
    
    def ResetGearSet(self):
        self.GearSet = {}

    def getWeaponDelay(self):
        """This function returns the weapon delay in seconds of the current weapon. If no weapon is equiped
           it returns the base value of 3.0
        """

        if "WEAPON" in self.GearSet.keys():
            return self.GearSet["WEAPON"].weaponDelay
        return 3.0
        
    def GetGearSetStat(self, IsTank : bool = False):
        self.IsTank = IsTank
        Stat = {
        "MainStat" : 450 if not self.IsTank else 410,
        "WD" : 0,
        "Det" : 390,
        "Ten" : 400,
        "SS" : 400,
        "SkS" : 400,
        "Crit" : 400,
        "DH" : 400,
        "Piety" : 390
    }       
        
                             # Getting stats from gear and materia
        for key in self.GearSet:
            GearPiece = self.GearSet[key]
            for type in StatType:
                Stat[type.name] += GearPiece.GetStat(type.name)

                             # Getting stats from food
        if self.Food != None:
            for type in StatType:
                percentBonus = self.Food.getPercentStatBonus(type.name)
                limitBonus = self.Food.getLimitStatBonus(type.name)

                if percentBonus != 0:
                    flatBonus = min(int(percentBonus * Stat[type.name]), limitBonus)
                    Stat[type.name] += flatBonus

        return Stat
    
    def hasValidMelding(self) -> bool:
        """
        This function returns weither all pieces in the GearSet have valid melding
        """
        valid = True
        for gear in self:
            valid = valid and gear.hasValidMelding()
        return valid
    
    def getMateriaTypeList(self, IgnoreSpeedMateria : bool = True,IgnorePietyMateria : bool = True, ignoreValidMeld : bool = False) -> list[StatType]:
        """
        This function returns a list of type of the materias present in the gear set
        IgnoreSpeedMateria : bool -> If true the function will only return non Speed related stattype. Default True
        IgnorePietyMateria : bool -> If true this function ignores Piety Materia. Default True
        ignoreValidMeld : bool -> If true, function only returns materias type that are on at least one gear with invalidMelding. Default False.
        """
        statTypeList = []
        for gear in self:
            if not (gear.hasValidMelding() and ignoreValidMeld) : statTypeList += [type for type in gear.getMateriaTypeList(IgnoreSpeedMateria=IgnoreSpeedMateria,IgnorePietyMateria=IgnorePietyMateria) if not (type in statTypeList)]
        return statTypeList    

def ImportGear(fileName : str) -> dict:
    """
    This function imports a list of gear. It reads the given file and will output a dictionnary with heach
    gear type as key with a list of all gear of that type in the import file.
    fileName : str -> Name of the file. Must be formatted correctly
    """
    GearDict = {}
    try:
        f = open(fileName) #Opening save
        GearDict = translateGear(json.load(f))
    except:
        raise InvalidFileName(fileName)

    return GearDict

def translateGear(data):
    """
    This function takes a list of gear and returns a dictionnary with each gear in a key corresponding to their
    type
    """

    GearDict = {}

    for GearPiece in data:
        type = GearType.name_for_id(GearPiece["GearType"])
        StatList = [Stat(StatType.id_for_name(S[0]), int(S[1])) for S in GearPiece["StatList"]]
        ImportedGear = Gear(GearPiece["GearType"], StatList, MateriaLimit = GearPiece["MateriaLimit"], Name = GearPiece["Name"])

                             # If is a weapon sets the weaponDelay to the desired value
        if "WeaponDelay" in GearPiece.keys(): ImportedGear.setWeaponDelay(float(GearPiece["WeaponDelay"]))

                             # If a custom stat limit is set uses this one instead of the highest stat one
        if "customStatLimit" in GearPiece.keys(): ImportedGear.setStatLimit(GearPiece["customStatLimit"])

        if type in GearDict.keys():
            GearDict[type].append(ImportedGear)
        else:
            GearDict[type] = [ImportedGear]

                            # Checking if ignoreOptimize
        if "ignoreOptimize" in GearPiece.keys():
                            # if true, then we also check for materias to put on and let the solver know to not optimize this gear piece
            ImportedGear.setIgnoreOptimize(True)
            if "defaultMateriaList" in GearPiece.keys():

                for materia in GearPiece["defaultMateriaList"]:
                            # materia will contain the type and the value. It will be assumed to be even.
                    matGen = MateriaGenerator(0,materia["value"])
                    ImportedGear.AddMateria(matGen.GenerateMateria(StatType(materia["type"])))

    return GearDict


#########################################
########## SAMURAI PLAYER ###############
#########################################

from ffxivcalc.Jobs.Melee.Melee_Spell import SamuraiSpell
from ffxivcalc.Jobs.Base_Spell import DOTSpell, buff, empty
import math
import copy
Lock = 0
#Special
def AddKenki(Player, Add):
    Player.KenkiGauge = min(100, Player.KenkiGauge + Add)
    #input("Kenki Gauge : " + str(Player.KenkiGauge))


#Requirement

def MeikyoRequirement(Player, Spell):
    return Player.MeikyoStack > 0, Player.MeikyoCD

def IkishotenRequirement(Player, Spell):
    return Player.IkishotenCD <= 0, Player.IkishotenCD

def KaeshiRequirement(Player, Spell):
    return Player.KaeshiCD <= 0, Player.KaeshiCD

def SeneiRequirement(Player, Spell):
    return Player.SeneiCD <= 0, Player.SeneiCD

def OgiNamikiriRequirement(Player, Spell):
    return Player.OgiNamikiriReady, -1

def KaeshiNamikiriRequirement(Player, Spell):
    return Player.KaeshiNamikiriReady, -1

def ShohaRequirement(Player, Spell):
    return Player.MeditationGauge == 3, -1

def MidareRequirement(Player, Spell):
    return Player.Setsu and Player.Ka and Player.Getsu, -1 #need all of them

def TenkaGokenRequirement(Player, Spell): #Need 2 
    i = 0
    if Player.Setsu : i +=1
    if Player.Ka : i+=1
    if Player.Getsu : i +=1
    return i == 2, -1

def HiganbanaRequirement(Player, Spell):#Need 1
    i = 0
    if Player.Setsu : i +=1
    if Player.Ka : i+=1
    if Player.Getsu : i +=1
    return i == 1, -1

def ThirdEyeRequirement(Player, Spell):
    return Player.ThirdEyeCD <= 0, Player.ThirdEyeCD

def GyotenRequirement(Player, Spell):
    return Player.GyotenCD <= 0, Player.GyotenCD

def YatenRequirement(Player, Spell):
    return Player.YatenCD <= 0, Player.YatenCD

def HagakureRequirement(Player, Spell):
    return Player.HagakureCD <= 0, Player.HagakureCD

def TsubamegaeshiRequirement(Player, Spell):
    return Player.TsubamegaeshiStack > 0, Player.TsubamegaeshiCD

def KaeshiHiganbanaRequirement(Player, Spell):
    return Player.KaeshiHiganbana, -1

def KaeshiGokenRequirement(Player, Spell):
    return Player.KaeshiGoken, -1

def KaeshiSetsugekkaRequirement(Player, Spell):
    return Player.KaeshiSetsugekka, -1

#Apply

def ApplyTsubamegaeshi(Player, Enemy):
    Player.KaeshiHiganbana , Player.KaeshiGoken, Player.KaeshiSetsugekka = False,False,False #Reseting values
    if Player.TsubamegaeshiStack == 2:
        Player.EffectCDList.append(TsubamegaeshiStackCheck)
        Player.TsubamegaeshiCD = 60
    Player.TsubamegaeshiStack -= 1

def ApplyHagakure(Player, Enemy):
    i = 0
    if Player.Setsu : i +=1
    if Player.Ka : i+=1
    if Player.Getsu : i +=1

    AddKenki(Player, i * 10)
    Player.Setsu, Player.Ka, Player.Getsu = False, False, False #Reseting

def ApplyYaten(Player, Enemy):
    Player.YatenCD = 10
    if not (YatenEffect in Player.EffectList) : 
        Player.EffectList.append(YatenEffect)
        Player.EffectCDList.append(YatenCheck)
    Player.EnhancedEnpiTimer = 15

def ApplyGyoten(Player, Enemy):
    Player.GyotenCD = 10

def Add10Kenki(Player, Enemy):
    AddKenki(Player, 10)

def ApplyTenkaGoken(Player, Enemy):
    Player.KaeshiGoken = True
    Player.EffectList.append(TsubamegaeshiCheck)
    Player.MeditationGauge = min(3, Player.MeditationGauge + 1) #Adding meditation stack
    Player.Setsu, Player.Ka, Player.Getsu = False, False, False #Reseting

def ApplyThirdEye(Player, Enemy):
    Player.ThirdEyeCD = 15
    #We will assume it gives us 10 Kenki
    AddKenki(Player, 10)

def ApplyFuko(Player, Enemy):
    AddKenki(Player, 10)
    if not (FukoCombo in Player.EffectList) : Player.EffectList.append(FukoCombo)

def ApplyHakaze(Player, Enemy):
    AddKenki(Player, 5)
    if not (HakazeEffect in Player.EffectList) : 
        Player.EffectList.append(HakazeEffect)
        #input("COMBO EFFECT")

def ApplyJinpu(Player, Enemy):
    pass

def ApplyShifu(Player, Enemy):
    if not (ShifuEffect in Player.EffectList) : Player.EffectList.append(ShifuEffect)

def ApplyMeikyo(Player, Enemy):
    if Player.MeikyoStack == 2:
        Player.EffectCDList.append(MeikyoStackCheck)
        Player.MeikyoCD = 55
    Player.MeikyoStack -= 1
    Player.EffectList.append(MeikyoEffect)
    Player.EffectCDList.append(MeikyoCheck) #Could be a problem if do it before finishing 3 weaponskills
    Player.Meikyo = 3


def ApplyIkishoten(Player, Enemy):
    AddKenki(Player, 50)
    Player.IkishotenCD = 120
    Player.OgiNamikiriReady = True

def ApplyMidare(Player, Enemy):
    Player.KaeshiSetsugekka = True
    Player.EffectList.append(TsubamegaeshiCheck)
    Player.MeditationGauge = min(3, Player.MeditationGauge + 1)
    Player.DirectCrit = True
    Player.Setsu, Player.Ka, Player.Getsu = False, False, False #Reseting

def ApplyKaeshi(Player, Enemy):
    Player.DirectCrit = True
    ApplyTsubamegaeshi(Player, Enemy) #Also sets flag to false

def ApplyKaeshiHiganbana(Player, Enemy):
    Player.DirectCrit = True
    ApplyTsubamegaeshi(Player, Enemy) #Also sets flag to false
    if Player.Higanbana == None:
        Player.Higanbana = copy.deepcopy(HiganbanaDOT)
        Player.DOTList.append(Player.Higanbana)
    else: Player.Higanbana.resetBuffSnapshot() # If already applied reset snapshot
    Player.HiganbanaTimer = 60 #Adding dot

def ApplyKaeshiGoken(Player, Enemy):
    Player.DirectCrit = True
    ApplyTsubamegaeshi(Player, Enemy) #Also sets flag to false


def ApplySenei(Player, Enemy):
    Player.SeneiCD = 120
    Player.KenkiGauge -= 25

def ApplyHiganbana(Player, Enemy):
    Player.KaeshiHiganbana = True
    Player.EffectList.append(TsubamegaeshiCheck)
    if Player.Higanbana == None:
        Player.Higanbana = copy.deepcopy(HiganbanaDOT)
        Player.DOTList.append(Player.Higanbana)
        Player.EffectCDList.append(HiganbanaCheck)
    else: Player.Higanbana.resetBuffSnapshot() # If already applied reset snapshot
    Player.HiganbanaTimer = 60
    Player.MeditationGauge = min(3, Player.MeditationGauge + 1)
    Player.Setsu, Player.Ka, Player.Getsu = False, False, False #Reseting

def ApplyOgiNamikiri(Player, Enemy):
    #print("hey")
    Player.KaeshiNamikiriReady = True
    Player.OgiNamikiriReady = False
    Player.DirectCrit = True
    Player.MeditationGauge = min(3, Player.MeditationGauge + 1)

def ApplyKaeshiNamikiri(Player, Enemy):
    Player.KaeshiNamikiriReady = False
    Player.DirectCrit = True

def ApplyShoha(Player, Enemy):
    Player.MeditationGauge = 0

def ApplyShinten(Player, Enemy):
    Player.KenkiGauge -= 25

def ApplyKaeshiGoken(Player, Enemy):
    ApplyTsubamegaeshi(Player, Enemy)

#Combo Action Effect

def FukoCombo(Player, Spell):
    if Spell.id == Mangetsu.id:
        Spell.Potency += 20
        Player.FugetsuTimer = 40
        Player.Getsu = True
    elif Spell.id == Oka.id:
        Spell.Potency += 20
        Player.FukaTimer = 40
        Player.Ka = True

def HakazeEffect(Player, Spell):
    if Spell.id == Jinpu.id:
        Player.FugetsuTimer = 40
        AddKenki(Player, 5)
        Spell.Potency += 160
        if not Player.Fugetsu:
            Player.Fugetsu = True
            Player.buffList.append(FugetsuBuff)
            Player.EffectCDList.append(FugetsuCheck)
            Player.EffectToRemove.append(HakazeEffect)
        if not (JinpuEffect in Player.EffectList) : Player.EffectList.append(JinpuEffect)
    elif Spell.id == Shifu.id:
        Spell.Potency += 160
        if not (FukaEffect in Player.EffectList) : Player.EffectList.append(FukaEffect)
        if not (FukaCheck in Player.EffectCDList) : 
            Player.Haste += 13
            Player.hasteHasChanged = True
            Player.hasteChangeValue = 13
            Player.EffectCDList.append(FukaCheck)
        Player.FukaTimer = 40
        AddKenki(Player, 5)
        Player.EffectToRemove.append(HakazeEffect)
        ApplyShifu(Player, Spell)
    elif Spell.id == Yukikaze.id:
        Spell.Potency += 180
        AddKenki(Player, 15)
        Player.Setsu = True
        Player.EffectToRemove.append(HakazeEffect)


def ShifuEffect(Player, Spell):
    if Spell.id == Kasha.id:
        Spell.Potency += 210
        AddKenki(Player, 10)
        Player.Ka = True
        Player.EffectToRemove.append(ShifuEffect)

def JinpuEffect(Player, Spell):
    if Spell.id == Gekko.id: 
        Spell.Potency +=210
        AddKenki(Player, 10)
        Player.Getsu = True
        Player.EffectToRemove.append(JinpuEffect)

def FukaEffect(Player, Spell):
    pass
    #Spell.CastTime *= 0.87
    #Spell.RecastTime *= 0.87

def MeikyoEffect(Player, Spell):
    if Spell.id == Gekko.id: 
        Spell.Potency +=210
        AddKenki(Player, 10)
        Player.Getsu = True
        Player.Meikyo -= 1
        Player.FugetsuTimer = 40
        if not Player.Fugetsu:
            Player.Fugetsu = True
            Player.buffList.append(FugetsuBuff)
            Player.EffectCDList.append(FugetsuCheck)
    elif Spell.id == Kasha.id:
        Spell.Potency += 210
        AddKenki(Player, 10)
        Player.Ka = True
        Player.Meikyo -= 1
        if not (FukaEffect in Player.EffectList) : Player.EffectList.append(FukaEffect)
        if not (FukaCheck in Player.EffectCDList) : 
            Player.Haste += 13
            Player.hasteHasChanged = True
            Player.hasteChangeValue = 13
            Player.EffectCDList.append(FukaCheck)
        Player.FukaTimer = 40
    elif Spell.id == Jinpu.id:
        Player.FugetsuTimer = 40
        AddKenki(Player, 5)
        Spell.Potency += 160
        if not Player.Fugetsu:
            Player.Fugetsu = True
            Player.buffList.append(FugetsuBuff)
            Player.EffectCDList.append(FugetsuCheck)
        Player.Meikyo -= 1
    elif Spell.id == Shifu.id:
        Spell.Potency += 160
        if not (FukaEffect in Player.EffectList) : Player.EffectList.append(FukaEffect)
        if not (FukaCheck in Player.EffectCDList) : 
            Player.Haste += 13
            Player.hasteHasChanged = True
            Player.hasteChangeValue = 13
            Player.EffectCDList.append(FukaCheck)
        Player.FukaTimer = 40
        AddKenki(Player, 5)
        Player.Meikyo -= 1
    elif Spell.id == Yukikaze.id:
        Spell.Potency += 180
        AddKenki(Player, 15)
        Player.Setsu = True
        Player.Meikyo -= 1

#Effect

def TsubamegaeshiCheck(Player, Spell):
    #This effect will remove the ability to do Tsubamegaeshi if the next spell after a Iaijutsu is not a Kaeshi
    if Spell.GCD:
        if Player.KaeshiHiganbana:
            if Spell.id != KaeshiHiganbana.id: Player.KaeshiHiganbana = False
        elif Player.KaeshiGoken:
            if Spell.id != KaeshiGoken.id: Player.KaeshiGoken = False
        elif Player.KaeshiSetsugekka:
            if Spell.id != KaeshiSetsugekka.id: Player.KaeshiSetsugekka = False

        Player.EffectToRemove.append(TsubamegaeshiCheck)

def YatenEffect(Player, Spell):
    if Spell.id == Enpi.id:
        Spell.Potency += 160
        Player.EnhancedEnpiTimer = 0 #End effect
    
#Effect Check

def YatenCheck(Player, Enemy):
    if Player.EnhancedEnpiTimer <= 0:
        Player.EffectList.remove(YatenEffect)
        Player.EffectToRemove.append(YatenCheck)

def FukaCheck(Player, Enemy):
    if Player.FukaTimer <= 0:
        Player.EffectList.remove(FukaEffect)
        Player.EffectToRemove.append(FukaCheck)
        Player.hasteHasChanged = True
        Player.Haste -= 13
        Player.hasteChangeValue = -13

def FugetsuCheck(Player, Enemy):
    if Player.FugetsuTimer <= 0:
        Player.Fugetsu = False
        Player.buffList.remove(FugetsuBuff)
        Player.EffectToRemove.append(FugetsuCheck)

def HiganbanaCheck(Player, Enemy):
    if Player.HiganbanaTimer <= 0:
        Player.DOTList.remove(Player.Higanbana)
        Player.Higanbana = None
        Player.EffectToRemove.append(HiganbanaCheck)

def MeikyoCheck(Player, Enemy):
    if Player.Meikyo == 0:
        Player.EffectList.remove(MeikyoEffect)
        Player.EffectToRemove.append(MeikyoCheck)

#Stack Check

def TsubamegaeshiStackCheck(Player, Enemy):
    if Player.TsubamegaeshiCD <= 0:
        if Player.TsubamegaeshiStack == 1:
            Player.EffectToRemove.append(TsubamegaeshiStackCheck)
        else:
            Player.TsubamegaeshiCD = 60
        Player.TsubamegaeshiStack += 1

def MeikyoStackCheck(Player, Enemy):
    if Player.MeikyoCD <= 0:
        if Player.MeikyoStack == 1:
            Player.EffectToRemove.append(MeikyoStackCheck)
        else:
            Player.MeikyoCD = 55
        Player.MeikyoStack +=1


#GCD
OgiNamikiri = SamuraiSpell(25781, True, 1.8, 2.5, 860, ApplyOgiNamikiri, [OgiNamikiriRequirement], 0, type = 2)
KaeshiNamikiri = SamuraiSpell(25782, True, 0, 1, 860, ApplyKaeshiNamikiri, [KaeshiNamikiriRequirement], 0)
Enpi = SamuraiSpell(7486, True, 0, 2.5, 100, Add10Kenki, [], 0, type = 2)

#Iaijutsu
Higanbana = SamuraiSpell(7489, True, 1.3, 2.5, 200, ApplyHiganbana, [HiganbanaRequirement], 0, type = 2)
HiganbanaDOT = DOTSpell(-1, 45, True)
TenkaGoken = SamuraiSpell(7488, True, 1.3, 2.5, 300, ApplyTenkaGoken, [TenkaGokenRequirement], 0, type = 2)
Midare = SamuraiSpell(7487, True, 1.3, 2.5, 640, ApplyMidare, [MidareRequirement], 0, type = 2)

#Kaeshi
KaeshiHiganbana = SamuraiSpell(16484, True, 0, 2.5, 200, ApplyKaeshiHiganbana, [KaeshiHiganbanaRequirement,TsubamegaeshiRequirement], 0, type = 2)
KaeshiGoken = SamuraiSpell(16485, True, 0, 2.5, 280, ApplyKaeshiGoken, [KaeshiGokenRequirement,TsubamegaeshiRequirement], 0, type = 2)
KaeshiSetsugekka = SamuraiSpell(16486, True, 0, 2.5, 640, ApplyKaeshi, [KaeshiSetsugekkaRequirement,TsubamegaeshiRequirement], 0, type = 2)

#Combo Actions
Hakaze = SamuraiSpell(7477, True, Lock, 2.5, 200, ApplyHakaze, [], 0 , type = 2)
Jinpu = SamuraiSpell(7478, True, Lock, 2.5, 120, empty, [], 0, type = 2)
Gekko = SamuraiSpell(7481, True, Lock, 2.5, 170, empty, [], 0, type = 2)
Shifu = SamuraiSpell(7479, True, Lock, 2.5, 120, empty, [], 0, type = 2)
Kasha = SamuraiSpell(7482, True, Lock, 2.5, 170, empty, [], 0, type = 2)
Yukikaze = SamuraiSpell(7480, True, Lock, 2.5, 120, empty, [], 0, type = 2)

#AOE Combo Action
Fuko = SamuraiSpell(25780, True, 0, 2.5, 100, ApplyFuko, [], 0 , type = 2)
Mangetsu = SamuraiSpell(7484, True, 0, 2.5, 100, empty, [], 0, type = 2)
Oka = SamuraiSpell(7485, True, 0, 2.5, 100, empty, [], 0, type = 2)

#oGCD
Meikyo = SamuraiSpell(7499, False, Lock, Lock, 0, ApplyMeikyo, [MeikyoRequirement], 0)
Ikishoten = SamuraiSpell(16482, False, Lock, Lock, 0, ApplyIkishoten, [IkishotenRequirement], 0)
Shoha = SamuraiSpell(16487, False, Lock, Lock, 560, ApplyShoha, [ShohaRequirement], 0)
Shoha2 = SamuraiSpell(25779, False, Lock, Lock, 200, ApplyShoha, [ShohaRequirement], 0) #AOEVersion of Shoha
ThirdEye = SamuraiSpell(7498, False, 0, 0, 0, ApplyThirdEye, [ThirdEyeRequirement], 0)
Hagakure = SamuraiSpell(7495, False, 0, 0, 0, ApplyHagakure, [HagakureRequirement], 0)

#Kenki Action
Senei = SamuraiSpell(16481, False, Lock, 2.5, 860, ApplySenei, [SeneiRequirement], 25)
Guren = SamuraiSpell(7496, False, 0, 0, 500, ApplySenei, [SeneiRequirement], 25) #AOE version of Senei
Shinten = SamuraiSpell(7490, False, Lock, 1, 250, ApplyShinten, [], 25)
Gyoten = SamuraiSpell(7492, False, 0, 0, 100, ApplyGyoten,[GyotenRequirement],10 )
Yaten = SamuraiSpell(7493, False, 0, 0, 100, ApplyYaten, [YatenRequirement], 10)
Kyuten = SamuraiSpell(7491, False, 0, 0, 110, empty, [], 25)
#buff
FugetsuBuff = buff(1.13,name="Hakaze")

def Meditate(time):
    #This function will return a Samurai Spell object that will correspond to a Meditate of the specified amount of time
    def MeditateRequirement(Player, Spell):
        return Player.MeditateCD <= 0 and time < 15 and time > 0, Player.MeditateCD

    def ApplyMeditate(Player, Spell):
        #Meditate generates 10 kenki each 3 sec, so we will divide time by 3, and add that much kenki
        AddKenki(Player, 10 * math.floor(time/3))
        Player.MeditateCD = 60

    return SamuraiSpell(7497, False, time, time, 0, ApplyMeditate, [MeditateRequirement], 0)

SamuraiAbility = {
7477 : Hakaze,
7479 : Shifu,
7482 : Kasha,
7478 : Jinpu,
7481 : Gekko,
7480 : Yukikaze,
25780 : Fuko,
7485 : Oka,
7484 : Mangetsu,
7499 : Meikyo,
7489 : Higanbana,
7488 : TenkaGoken,
7487 : Midare,
25781 : OgiNamikiri,
16484 : KaeshiHiganbana,
16485 : KaeshiGoken,
16486 : KaeshiSetsugekka,
25782 : KaeshiNamikiri,
7490 : Shinten,
7491 : Kyuten,
16481 : Senei,
7496 : Guren,
16487 : Shoha,
25779 : Shoha2,
16482 : Ikishoten,
7497 : Meditate(3),
7495 : Hagakure,
7492 : Gyoten,
7493 : Yaten,
7486 : Enpi,
7498 : ThirdEye


}


import os

from ffxivcalc.Enemy import Enemy
from ffxivcalc.Fight import Fight
from ffxivcalc.Jobs.Player import Player
from ffxivcalc.Jobs.PlayerEnum import *
from copy import deepcopy

from ffxivcalc.Jobs.Base_Spell import WaitAbility, Potion
from ffxivcalc.Jobs.Caster.Caster_Spell import *
from ffxivcalc.Jobs.Melee.Melee_Spell import *
from ffxivcalc.Jobs.Ranged.Ranged_Spell import *
from ffxivcalc.Jobs.Healer.Healer_Spell import *
from ffxivcalc.Jobs.Tank.Tank_Spell import *

#CASTER
from ffxivcalc.Jobs.Caster.Summoner.Summoner_Spell import *
from ffxivcalc.Jobs.Caster.Blackmage.BlackMage_Spell import * 
from ffxivcalc.Jobs.Caster.Redmage.Redmage_Spell import *

#HEALER
from ffxivcalc.Jobs.Healer.Sage.Sage_Spell import *
from ffxivcalc.Jobs.Healer.Scholar.Scholar_Spell import *
from ffxivcalc.Jobs.Healer.Whitemage.Whitemage_Spell import *
from ffxivcalc.Jobs.Healer.Astrologian.Astrologian_Spell import *

#RANGED
from ffxivcalc.Jobs.Ranged.Machinist.Machinist_Spell import *
from ffxivcalc.Jobs.Ranged.Bard.Bard_Spell import *
from ffxivcalc.Jobs.Ranged.Dancer.Dancer_Spell import *

#TANK
from ffxivcalc.Jobs.Tank.Gunbreaker.Gunbreaker_Spell import *
from ffxivcalc.Jobs.Tank.DarkKnight.DarkKnight_Spell import *
from ffxivcalc.Jobs.Tank.Warrior.Warrior_Spell import *
from ffxivcalc.Jobs.Tank.Paladin.Paladin_Spell import *

#MELEE
from ffxivcalc.Jobs.Melee.Samurai.Samurai_Spell import *
from ffxivcalc.Jobs.Melee.Ninja.Ninja_Spell import *
from ffxivcalc.Jobs.Melee.Dragoon.Dragoon_Spell import *
from ffxivcalc.Jobs.Melee.Reaper.Reaper_Spell import *
from ffxivcalc.Jobs.Melee.Monk.Monk_Spell import *

from ffxivcalc.helperCode.helper_backend import SaveFight

def ExecuteMemoryCode(SaveFight_check, time_unit=0.01, TimeLimit=500):
    # This part of the code will execute whatever rotation is written here. It will be called from TUI.

    Dummy = Enemy()
    Event = Fight(Dummy, False)


    # ===============================================================================================
    # You don't need to to worry about anything above this point

    # Stat Sheet
    # Enter your own stats here. The default is the 6.2 Savage BiS found on the Balance. The default Tenacity for non-tank is 400.
    # These stats must include the bonus stats food gives.

    # Caster
    BLMStat = {"MainStat": 2945, "WD":126, "Det" : 1451, "Ten" : 400, "SS": 840, "SkS" : 400,  "Crit" : 2386, "DH" : 1307} # Stats for BlackMage
    RDMStat = {"MainStat": 2947, "WD":126, "Det" : 1548, "Ten" : 400, "SS": 495, "SkS" : 400, "Crit" : 2397, "DH" : 1544} # Stats for RedMage
    SMNStat = {"MainStat": 2948, "WD":126, "Det" : 1451, "Ten" : 400, "SS": 544, "SkS" : 400, "Crit" : 2436, "DH" : 1544} # Stats for Summoner

    # Healer
    SCHStat = {"MainStat": 2931, "WD":126, "Det" : 1750, "Ten" : 400, "SS": 1473, "SkS" : 400, "Crit" : 2351, "DH" : 436} # Stats for Scholar
    WHMStat = {"MainStat": 2945, "WD":126, "Det" : 1792, "Ten" : 400, "SS": 839, "SkS" : 400, "Crit" : 2313, "DH" : 904} # Stats for WhiteMage
    ASTStat = {"MainStat": 2949, "WD":126, "Det" : 1659, "Ten" : 400, "SS": 1473, "SkS" : 400, "Crit" : 2280, "DH" : 436} # Stats for Astrologian
    SGEStat = {"MainStat": 2928, "WD":126, "Det" : 1859, "Ten" : 400, "SS": 827, "SkS" : 400, "Crit" : 2312, "DH" : 1012} # Stats for Sage

    # Physical Ranged
    MCHStat = {"MainStat": 2937, "WD":126, "Det" : 1598, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2389, "DH" : 1592} # Stats for Machinist
    BRDStat = {"MainStat": 2949, "WD":126, "Det" : 1721, "Ten" : 400, "SS": 400, "SkS" : 536, "Crit" : 2387, "DH" : 1340} # Stats for Bard
    DNCStat = {"MainStat": 2949, "WD":126, "Det" : 1721, "Ten" : 400, "SS": 400, "SkS" : 536, "Crit" : 2387, "DH" : 1340} # Stats for Dancer

    # Melee
    NINStat = {"MainStat": 2921, "WD":126, "Det" : 1669, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2399, "DH" : 1511} # Stats for Ninja
    SAMStat = {"MainStat": 2937, "WD":126, "Det" : 1571, "Ten" : 400, "SS": 400, "SkS" : 508, "Crit" : 2446, "DH" : 1459} # Stats for Samurai
    DRGStat = {"MainStat": 2949, "WD":126, "Det" : 1545, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2462, "DH" : 1577} # Stats for Dragoon
    MNKStat = {"MainStat": 3076, "WD":126, "Det" : 1546, "Ten" : 400, "SS": 400, "SkS" : 769, "Crit" : 2490, "DH" : 1179} # Stats for Monk
    RPRStat = {"MainStat": 2946, "WD":126, "Det" : 1545, "Ten" : 400, "SS": 400, "SkS" : 400, "Crit" : 2462, "DH" : 1577} # Stats for Reaper

    # Tank
    DRKStat = {"MainStat": 2910, "WD":126, "Det" : 1844, "Ten" : 751, "SS": 400, "SkS" : 400, "Crit" : 2377, "DH" : 1012} # Stats for DarkKnight
    WARStat = {"MainStat": 2910, "WD":126, "Det" : 1844, "Ten" : 751, "SS": 400, "SkS" : 400, "Crit" : 2377, "DH" : 1012} # Stats for Warrior
    PLDStat = {"MainStat": 2891, "WD":126, "Det" : 1883, "Ten" : 631, "SS": 400, "SkS" : 650, "Crit" : 2352, "DH" : 868} # Stats for Paladin
    GNBStat = {"MainStat": 2891, "WD":126, "Det" : 1883, "Ten" : 631, "SS": 400, "SkS" : 650, "Crit" : 2352, "DH" : 868} # Stats for Gunbreaker

    # ===============================================================================================

    # Here the player objects are being initialized. You do not need to change anything here.
    # Note that if you want to simulate with two players of the same Jobs you will need to create another Player Object here.
    # You can simply copy the objet's __init__ and change the name. 

    # Caster player object
    BLMPlayer = Player([], [], deepcopy(BLMStat), JobEnum.BlackMage)
    RDMPlayer = Player([], [], deepcopy(RDMStat), JobEnum.RedMage)
    SMNPlayer = Player([], [], deepcopy(SMNStat), JobEnum.Summoner)

    # Healer player object
    SCHPlayer = Player([], [], deepcopy(SCHStat), JobEnum.Scholar)
    WHMPlayer = Player([], [], deepcopy(WHMStat), JobEnum.WhiteMage)
    SGEPlayer = Player([], [], deepcopy(SGEStat), JobEnum.Sage)
    ASTPlayer = Player([], [], deepcopy(ASTStat), JobEnum.Astrologian)

    # Physical Ranged
    MCHPlayer = Player([], [], deepcopy(MCHStat), JobEnum.Machinist)
    BRDPlayer = Player([], [], deepcopy(BRDStat), JobEnum.Bard)
    DNCPlayer = Player([], [], deepcopy(DNCStat), JobEnum.Dancer)
    
    # Melee
    NINPlayer = Player([], [], deepcopy(NINStat), JobEnum.Ninja)
    SAMPlayer = Player([], [], deepcopy(SAMStat), JobEnum.Samurai)
    DRGPlayer = Player([], [], deepcopy(DRGStat), JobEnum.Dragoon)
    RPRPlayer = Player([], [], deepcopy(RPRStat), JobEnum.Reaper)
    MNKPlayer = Player([], [], deepcopy(MNKStat), JobEnum.Monk)

    # Tank
    DRKPlayer = Player([], [], deepcopy(DRKStat), JobEnum.DarkKnight)
    WARPlayer = Player([], [], deepcopy(WARStat), JobEnum.Warrior)
    PLDPlayer = Player([], [], deepcopy(PLDStat), JobEnum.Paladin)
    GNBPlayer = Player([], [], deepcopy(GNBStat), JobEnum.Gunbreaker)


    # ===============================================================================================

    # Here you can enter the action list you want the simulator to simulate. If you want to simulate a BlackMage go to the respective list, in that case you would write in BLMOpener\
    # Note that this action list also includes the prepull. The simulator will start at soon as one of the player character does damage. So coordinate your different player with WaitAbility() so
    # they all start at the same time.
    # Note that if you are simulating with more than 1 per job you will need to create a new list of actions.

    # Caster
    BLMOpener = [Fire3, Thunder3, Fire4, Triplecast, Potion, Fire4, Amplifier, LeyLines, Fire4, SharpCast, Swiftcast, Despair, Triplecast, Manafront, Fire4, Despair]
    SMNOpener = [Ruin3, Summon, SearingLight, AstralImpulse, AstralImpulse, AstralImpulse, EnergyDrainSMN, Enkindle, AstralImpulse, Deathflare, Fester, AstralImpulse, Fester, AstralImpulse, Garuda, Swiftcast, Slipstream, Emerald, Emerald, Emerald, Emerald, Titan]
    RDMOpener = [Verthunder, Verareo, Swiftcast,Acceleration, Verthunder, Verthunder, Embolden, Manafication, EnchantedRiposte, Fleche, EnchantedZwerchhau, Contre, EnchantedRedoublement, Corps, Engagement, Verholy, Corps, Engagement, Scorch, Resolution, Verfire, Verthunder, Verstone, Verareo, Jolt, Verthunder, Fleche]
    
    # Healer
    SCHOpener = [Broil, Biolysis, Aetherflow, Broil, Swiftcast, Broil, ChainStratagem, EnergyDrain, Broil, EnergyDrain, Broil, EnergyDrain, Broil, Dissipation, Broil, EnergyDrain, Broil, EnergyDrain, Broil, EnergyDrain]
    WHMOpener = [Glare, Dia, Glare, Glare, PresenceOfMind, Glare, Assize, Glare, Glare, Glare, Glare, Glare, Glare, Glare, Glare, Glare, Glare, Glare]
    ASTOpener = [Malefic, Lightspeed, Combust, Malefic, Malefic, Divination, Malefic, MinorArcana, Astrodyne, Malefic, LordOfCrown, Malefic ]
    SGEOpener = [Dosis, Eukrasia, EukrasianDosis, Dosis, Dosis, Phlegma, Phlegma, Dosis, Dosis, Dosis, Dosis, Dosis, Dosis, Dosis, Dosis, Dosis, Dosis]

    # Physical Ranged 
    BRDOpener = [Stormbite, WanderingMinuet, RagingStrike, Causticbite, EmpyrealArrow, BloodLetter, RefulgentArrow, RadiantFinale, BattleVoice, BurstShot, Barrage, RefulgentArrow, Sidewinder, BurstShot, RefulgentArrow, BurstShot, EmpyrealArrow, IronJaws, PitchPerfect3, WaitAbility(50)]
    MCHOpener = [Reassemble, WaitAbility(5), AirAnchor, GaussRound, Ricochet, Drill, BarrelStabilizer, SplitShot, SlugShot, GaussRound, Ricochet, CleanShot, Reassemble, WaitAbility(2.2), Wildfire, ChainSaw, Automaton, WaitAbility(2),Hypercharge, HeatBlast, Ricochet, HeatBlast, GaussRound, HeatBlast, Ricochet, HeatBlast, GaussRound, HeatBlast, Ricochet, Drill, WaitAbility(60)]
    DNCOpener = [StandardStep, Pirouette, Jete, WaitAbility(15), StandardFinish, TechnicalStep, Pirouette, Jete, Entrechat, Emboite, TechnicalFinish, Devilment, StarfallDance, Flourish, FanDance3, Tillana, FanDance4, FountainFall, FanDance1, FanDance3, StandardStep, Jete, Pirouette, StandardFinish]

    # Melee
    SAMOpener = [Meikyo, WaitAbility(9), Gekko, Kasha, Ikishoten, Yukikaze, Midare, Senei, KaeshiSetsugekka, Meikyo, Gekko, Shinten, Higanbana,Shinten, OgiNamikiri, Shoha, KaeshiNamikiri, Kasha, Shinten, Gekko, Gyoten, Hakaze, Yukikaze, Shinten, Midare, KaeshiSetsugekka ]
    DRGOpener = [TrueThrust, Disembowel, LanceCharge, DragonSight(NINPlayer), ChaoticSpring, BattleLitany, WheelingThrust, Geirskogul, LifeSurge, FangAndClaw, HighJump, RaidenThrust, DragonFireDive, VorpalThrust, LifeSurge, MirageDive, HeavenThrust, SpineshafterDive, FangAndClaw, SpineshafterDive, WheelingThrust, RaidenThrust, WyrmwindThrust, Disembowel, ChaoticSpring, WheelingThrust, WaitAbility(50)]
    MNKOpener = [FormShift, DragonKick, TwinSnakes, RiddleOfFire, Demolish, TheForbiddenChakra, Bootshine, Brotherhood, PerfectBalance, DragonKick, RiddleOfWind, Bootshine, DragonKick, ElixirField, Bootshine, PerfectBalance, TwinSnakes, DragonKick, Demolish, RisingPhoenix]
    NINOpener = [Jin, Chi, Ten, Huton, Hide, Ten, Chi, Jin, WaitAbility(5), Suiton, Kassatsu, SpinningEdge, GustSlash, Mug, Bunshin, PhantomKamaitachi, TrickAttack, AeolianEdge, DreamWithinADream, Ten, Jin, HyoshoRanryu, Ten, Chi, Raiton, TenChiJin, Ten2, Chi2, Jin2, Meisui, FleetingRaiju, Bhavacakra, FleetingRaiju, Bhavacakra, Ten, Chi, Raiton, FleetingRaiju ]
    RPROpener = [Harpe, ShadowOfDeath, ArcaneCircle, SoulSlice, SoulSlice, PlentifulHarvest, Enshroud, VoidReaping, CrossReaping, LemureSlice, VoidReaping, CrossReaping, LemureSlice, Communio, Gluttony, Gibbet, Gallows, UnveiledGibbet, Gibbet]

    # Tank 
    DRKOpener = [BloodWeapon, WaitAbility(4), HardSlash, EdgeShadow, Delirium, SyphonStrike, Souleater, LivingShadow, SaltedEarth, HardSlash, Shadowbringer, EdgeShadow, Bloodspiller, CarveSpit, Plunge, Bloodspiller, Shadowbringer, EdgeShadow, Bloodspiller, SaltDarkness, EdgeShadow, SyphonStrike, Plunge, EdgeShadow]
    WAROpener = [Tomahawk, Infuriate, HeavySwing, Maim, StormEye, InnerRelease, InnerChaos, Upheaval, Onslaught, PrimalRend, Infuriate, InnerChaos, Onslaught, FellCleave, Onslaught, FellCleave, FellCleave, HeavySwing, Maim, StormPath, FellCleave, Infuriate, InnerChaos]
    PLDOpener = [HolySpirit, FastBlade, FightOrFlight, RiotBlade, GoringBlade, FastBlade, RiotBlade, CircleScorn, Expiacion, RoyalAuthority, Intervene, RequestACat, Atonement, Intervene, Atonement, Atonement, FastBlade, RiotBlade, GoringBlade, HolySpirit, HolySpirit, HolySpirit, HolySpirit, Confetti,BladeFaith, BladeTruth, BladeValor, WaitAbility(50) ]
    GNBOpener = [LightningShot, KeenEdge, BrutalShell, NoMercy, Bloodfest, GnashingFang, JugularRip, SonicBreak, BlastingZone, BowShock, DoubleDown, RoughDivide, SavageClaw, AbdomenTear, RoughDivide, WickedTalon, EyeGouge, SolidBarrel, BurstStrike, Hypervelocity, KeenEdge, WaitAbility(50)]


    # ===============================================================================================

    # Here we are linking the earlier created action list to the player object. You should not have to change anything here except if you are trying to simulate
    # with more than 1 player per job. In which case you will need to link the earlier created object and the earlier created action list.

    # Caster
    BLMPlayer.ActionSet = BLMOpener
    RDMPlayer.ActionSet = RDMOpener
    SMNPlayer.ActionSet = SMNOpener

    # Healer
    SCHPlayer.ActionSet = SCHOpener
    WHMPlayer.ActionSet = WHMOpener
    ASTPlayer.ActionSet = ASTOpener
    SGEPlayer.ActionSet = SGEOpener

    # Physical Ranged
    MCHPlayer.ActionSet = MCHOpener
    BRDPlayer.ActionSet = BRDOpener
    DNCPlayer.ActionSet = DNCOpener

    # Melee
    NINPlayer.ActionSet = NINOpener
    SAMPlayer.ActionSet = SAMOpener
    DRGPlayer.ActionSet = DRGOpener
    RPRPlayer.ActionSet = RPROpener
    MNKPlayer.ActionSet = MNKOpener

    #Tank
    DRKPlayer.ActionSet = DRKOpener
    WARPlayer.ActionSet = WAROpener
    PLDPlayer.ActionSet = PLDOpener
    GNBPlayer.ActionSet = GNBOpener

    # ===============================================================================================

    # Here you will put into this list all the players you wish to simulate.
    # Note that the limit is not 8, you can put as much as you want.
    # Furthemore the simulator will compute the bonus 5% if it applies.
    # So if you want to simulate the BlackMage and a RedMage, you would do: 
    # PlayerList = [BLMPlayer, RDMPlayer]

    PlayerList = [BLMPlayer]
 
    Event.AddPlayer(PlayerList)

    # ===============================================================================================

    # Here you can change the final parameters to the simulation

    #TimeLimit = 50 # Time limit for the simulation in seconds. It will stop once this time has been reached (in simulation time)
    #time_unit = 0.01 # Time unit or frame of the simulation. Smallest step the simulator will take between each iterations. It is advised to not change this value
    ShowGraph = True # Parameter to show (or not) the graph generated by the simulator.
    RequirementOn = True # Parameter that will enable or disable the requirement check for all actions. If False the simulator will not check if an action can be done
    IgnoreMana = True # True if want to ignore mana

    # ===============================================================================================

    if SaveFight_check: #If we wish to save the fight

        os.system('CLS') #clearing HUD

        print(
        "===================== SAVE FIGHT IN CODE ====================="
        )
        filename = input("Enter the name you wish to save the fight as : ")

        SaveFight(Event, 0, TimeLimit, filename)
    else:
        Event.RequirementOn = RequirementOn
        Event.ShowGraph = ShowGraph
        Event.IgnoreMana = IgnoreMana
        Event.SimulateFight(time_unit, TimeLimit, vocal=True)
        print("=========================================================")
        input("Press any key to go back to the main")




class MateriaOverflow(Exception):
    """
    This exception is raised when trying to add a Materia to a gear that has reached its limit.
    """
    def __init__(self):
        self.message = "This gear piece cannot receive anymore Materias."

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'

class InvalidFunctionParameter(Exception):
    """
    This exception is raised when a function's parameter is invalid.
    """
    def __init__(self, funcName : str, paramName : str, info : str):
        self.message = "The parameter " + paramName + " in the function " + funcName + " is invalid. Info " + info

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class InvalidStatRequest(Exception):
    """
    This exception is raised when the user does an invalid request of stat of a gear piece
    """
    pass

class InvalidGearSpace(Exception):
    """
    This exception is raised when the GearSpace is invalid.
    """
    def __init__(self, missingKey : str):
        self.message = "The GearSpace is missing gear piece of type " + missingKey

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'

class MultiValuedWeaponDelay(Exception):
    """This exception is raised when weapons in a gear space have different weapon delay value
    """

    def __init__(self, weaponDelay : int, newWeaponDelay : int):
        self.message = f"The GearSpace has at least two weapons with different 'weaponDelay' value. Please make sure these values are identical : ( {weaponDelay} != {newWeaponDelay} )"

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'

class InvalidFoodSpace(Exception):
    """
    This exception is raised when the FoodSpace is invalid.
    """
    def __init__(self):
        self.message = "The FoodSpace cannot be empty."

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'

class InvalidMateriaSpace(Exception):
    """
    This exception is raised when the MateriaSpace is invalid.
    """
    def __init__(self):
        self.message = "The size of the MateriaSpace must be at least 3."

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
class InvalidTankBusterTargetNumber(Exception):
    """
    This exception is raised when an event is defined as a tank buster
    but has an invalid number of targets. Tank buster's number of targets
    must be 1 or 2. This error will be raised only if the Event has its "experimental"
    flag set to false (false by default). It must be, in any case, a positive value.

    Attributes:
        nTBTarget (int) : number of targets of the TB

    """

    def __init__(self, nTBTarget : int, id : int):
        self.nTBTarget = nTBTarget
        self.message = "targets for the tank buster with id " + str(id) + " is not in the valid range of 1 or 2."

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.nTBTarget} {self.message}'

class InvalidTarget(Exception):
    """
    This exception is raised when a non valid target is given as input to an action

    e.g: If an action cannot target the player casting it
         If the targetID is not an existing player

    Attributes:
        ActionName (str) : Name of the action which is given non valid target
        Caster (object) : The player object of the caster of the action.
        Target (object) : What was passed as target. Will display it as a string.
        InvalidID (bool) : If the given target ID is not a valid ID.
        TargetID (int) : ID of the target. Only used if InvalidID is True.
    """

    def __init__(self, ActionName : str, Caster, Target, InvalidID : bool, TargetID : int):
        self.Actionname = ActionName
        self.Target = str(Target)
        self.InvalidID = InvalidID
        self.message = "Player with ID "+str(Caster.playerID)+" has an invalid target. The given target for " + ActionName + " is not valid. Given target : "
        self.message_2 = "" if not InvalidID else " TargetID " + str(TargetID) + " is invalid." 

        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message} {self.Target} {self.message_2}'
    
    def __repr__(self):
        return str(self)


class InvalidMitigation(Exception):
    """This exception is raised when a non valid mitigation object is being constructed.
    Such examples would be MagicMit and PhysicalMit being both True or a MitPercent value out of the
    acceptable range of (0,1).

    Args:
        InvalidRange (bool): Type of the Error. True -> MitPercent OOR, False -> MagicMit and PhysicalMit == True
        PercentMit (float) : Value of the PercentMit given
    """

    def __init__(self, InvalidRange = False, PercentMit : float = 0):
        self.PercentMit = PercentMit

        self.message = "A mitigation that is both only for Physical and for Magic damage was being constructed" if not InvalidRange else "The given PercentMit value " + str(self.PercentMit) + " is not within the valid range of (0,1)."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class InvalidFileName(Exception):
    """This exception is raised when given an invalid file name

    fileName : str
    """
    def __init__(self, fileName : str):

        self.message = "Invalid file name" + fileName
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class ActionNotFound(Exception):#Exception called if an action isn't found in the dictionnary
    pass
class JobNotFound(Exception):#Exception called if a Job isn't found
    pass

class playerIDNotFound(Exception):#Exception called if the player isn't found with the ID given by using the Fight.playerForID function.
    pass

class emptyEventList(Exception):
    # Exception is called if trying to give an enemy an empty event list.
    pass

class invalidFFLogsFightId(Exception):
    # Exception is called when the users asks to retrieve a fight from a fight list with a wrong id.
    def __init__(self, fight_id : str):

        self.message = "The provided fight id is invalid. Provided id : " + fight_id
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class invalidFFLogsQuery(Exception):
    # Exception is called when the users does an invalid query to fflogs
    def __init__(self, message : str):

        self.message = "Invalid FFLogs query. The query returned with this error message : " + message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class outDatedFFLogsCode(Exception):
    # Exception is raised when the user imports the FFLogsAPIRequest module.
    def __init__(self):

        self.message = "The module 'ffxivcalc.Request.FFLogsAPIRequest' is no longer up to date. Use 'ffxivcalc.Request.FFLogs_api' instead."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class outDatedAPICode(Exception):
    # Expcetion is raised when the user tries to import the API module.
    def __init__(self):

        self.message = "The module 'ffxivcalc.API' is outdated and has been deactivated, please do not import it. If you still want to use it add a try/except block (not recommend)."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'
    
class deactivatedModuleCodeParser(Exception):
    # Expcetion is raised when the user tries to import the codeParser module.
    def __init__(self):

        self.message = "The module 'ffxivcalc.codeParser' is outdated and has been deactivated, please do not import it. If you still want to use it add a try/except block (not recommend)."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.message}'




from copy import deepcopy
from ffxivcalc.helperCode.helper_math import roundDown
from ffxivcalc.helperCode.exceptions import emptyEventList
# Exception
from ffxivcalc.helperCode.exceptions import InvalidTankBusterTargetNumber

class EnemyEvent:
    """
    This class represents an action or event the boss can take. Such events can be raidwide, a mechanic, untargetable, an enrage, etc.
    """

    def __init__(self, id, CastTime : float, Damage : int, RaidWide=True, nTBTarget=0, IsPhysical = False, Experimental = False) -> None:
        """Constructor the of the EnemyEvent class

        Args:
            id (int): id of the event.
            CastTime (float): casting time of the event.
            Damage (int): Damage the players receive when the action is used.
            RaidWide (bool) : True if the Event is a raidwide. Default value is true
            nTBTarget (int) : Number of targets of a tank buster. Only needed if RaidWide is false. Must be 1 or 2 if Experimental is false
            IsPhysical (bool) : True if the damage of that event is physical. By default false (hence magical by default)
            Experimental (bool) : False if we do not wish to overrule the automatic checking for a valid action.
        """

        if not Experimental and not RaidWide and (nTBTarget != 1 and nTBTarget != 2) or nTBTarget < 0 :
            # If a tankbuster but invalid number of targets and not experimental
            raise InvalidTankBusterTargetNumber(nTBTarget, id)

        self.CastTime = CastTime
        self.id = id
        self.Damage = Damage
        self.RaidWide = RaidWide
        self.nTBTarget = nTBTarget
        self.IsPhysical = IsPhysical
        self.target = [] # Empty list. The targets will be computed in begin_cast()


    def begin_cast(self, Enemy) -> None:
        """
        This function will begin the casting of an action by the enemy.

        Enemy (Enemy) : Enemy object casting the Event

        """

        Enemy.CastingTimer = self.CastTime
        Enemy.CastingAction = deepcopy(self)
        Enemy.IsCasting = True

    def cast(self, Enemy, RaidWide=True) -> None:
        """This function will cast and apply damage/effect on all the players

        Args:
            Enemy (Enemy): Enemy casting the action
        """

        # Finds targets

        self.target = Enemy.CurrentFight.PlayerList if self.RaidWide else Enemy.CurrentFight.GetEnemityList(self.nTBTarget)
        # If the event is a raidwide the target is all players.
        # If it is not it targets the players with the most enemity up to the number of targets
        # Will do damage on all target

        # Computes new damage because mitigation on the Enemy
        # Does not differentiate between magical and physical for now
        curr_mit = 1
        if Enemy.Addle > 0: 
            if self.IsPhysical : curr_mit = round(0.95 * curr_mit, 2) # 5% physical mit
            else : curr_mit = round(0.9 * curr_mit, 2) # 10% magic mit
        if Enemy.Feint  > 0: 
            if self.IsPhysical : curr_mit = round(0.9 * curr_mit, 2) # 10% physical mit
            else : curr_mit = round(0.95 * curr_mit, 2) # 5% magic mit
        if Enemy.Reprisal  > 0: curr_mit = round(0.9 * curr_mit, 2) # Flat 10% mit

        self.Damage *= curr_mit # Updating the new damage based on global mit



        for player in self.target:
            # Going through all players
            # Mit according to their own personnal mit
            if self.IsPhysical : player_damage = self.Damage * player.PhysicalMitigation # Only applies physical mit
            else : player_damage = self.Damage * player.MagicMitigation # Only applies magic mit
            player_damage = round(player_damage, 0) # Rounding down to lowest integer
            player.TakeDamage(player_damage, not self.IsPhysical) # Applying the damage to the player

        Enemy.EventNumber += 1 # Incrementing the pointer to the next event

        if Enemy.EventNumber == len(Enemy.EventList):
            # If the enemy has reached the end of its action list
            Enemy.hasEventList = False

class EnemyDOT(EnemyEvent):
    """
    This class is any DOT applied by the enemy on the players. It will do damage over time on the players.
    """

    def __init__(self, id, DOTDamage : int, DOTDuration):
        """
        Creates a DOTDamage object.

        DOTDamage (int) : Damage every application of the DOT
        DOTDuration (float) : Duration of the DOT.
        """
        self.id = id
        self.DOTDamage = DOTDamage
        self.DOTDuration = DOTDuration

        self.DOTStateTimer = 3 # The DOT will be applied every 3 seconds

    def updateState(self, time : float, player):
        """
        Update the states of the DOT. Applies damag if neeeded.

        time (float) : time by which the DOT is updated.
        player (Player) : player object on which the DOT is applied.

        """
        # Updating timers
        self.DOTStateTimer -= time
        self.DOTDuration -= time

        if self.DOTStateTimer <= 0: # Apply damage
            player.TakeDamage(self.DOTDamage)
            self.DOTStateTimer = 3

        if self.DOTDuration <= 0:
            # DOT is finished
            player.EnemyDOT.remove(self)

class Enemy:

    """
    This class is the enemy of the simulation. Contains some global effect that will apply to the entire group and keeps track of the total damage done
    """

    def __init__(self):


        self.EffectList = [] # List of all effect on the boss. These are effects applied by the players

        self.TotalPotency = 0
        self.TotalDamage = 0

        self.buffList = [] # List of all buffs the boss gives when a player attacks them

        # Buff that can be applied directly on the boss
        self.ChainStratagem = False # +10% crit rate
        self.WanderingMinuet = False # +2% crit rate
        self.BattleVoice = False # +20% direct hit
        self.ArmyPaeon = False # + 3% direct hit

        # Mitigation buff
        self.Addle = False # 10% magical, 5% physical
        self.Feint = False # 10% physical, 5% magical
        self.Reprisal = False # 10% true mitigation
        

        self.EventList = [] # List of all EnemyEvent object this boss will perform through the simulation
        self.EventNumber = 0 # Current index of the EvenList action.
        self.hasEventList = False # By default an enemy has nothing to do
        # Timer
        self.CastingTimer = 0
        self.AddleTimer = 0
        self.FeintTimer = 0
        self.ReprisalTimer = 0

        # Currently Casting Action
        self.CastingAction = None
        self.IsCasting = False # If the Enemy is casting

        # Fight the Enemy is in
        self.CurrentFight = None

    def setEventList(self, newEventList) -> None:
        """
        This function gives the Enemy an event list. This will result in a crash
        if the list is empty.

        Args:
            newEventList (List[EnemyEvent]): List of the Events to perform
        """

        if len(newEventList) == 0 : raise emptyEventList

        self.hasEventList = True # Let the Fight know this Enemy has an EventList
        
        self.EventList = deepcopy(newEventList)

    def UpdateTimer(self, time : float) -> None:
        """This function updates the Timer values of the Enemy object

        Args:
            time (float): value by which we update the timer
        """

        if self.AddleTimer > 0 : 
            self.AddleTimer = max(0, self.AddleTimer - time)
            if self.AddleTimer == 0 : self.Addle = False
        if self.FeintTimer > 0 : 
            self.FeintTimer = max(0, self.FeintTimer - time)
            if self.FeintTimer == 0 : self.Feint = False
        if self.ReprisalTimer > 0 : 
            self.ReprisalTimer = max(0, self.ReprisalTimer - time)
            if self.ReprisalTimer == 0 : self.Reprisal = False

        if self.IsCasting : # Only decrements if the enemy is currently casting
            self.CastingTimer = max(0, self.CastingTimer - time)

            if self.CastingTimer <= 0 : # If reaches the end of the cast
                self.CastingAction.cast(self) # Cast the action
                self.IsCasting = False

def WaitEvent(time : float) -> EnemyEvent:
    """
    This function returns a EnemyEvent object which has no effect or damage and simply makes the Enemy way.

    time (float) : time in seconds we want the enemy to wait
    """

    return EnemyEvent(-212,time, 0)

MagicRaidWide = EnemyEvent(1, 2, 400)
PhysicalRaidWide = EnemyEvent(3, 2, 500, IsPhysical=True)
TankBuster = EnemyEvent(2, 2, 1000, RaidWide=False, nTBTarget=1)



